"""RT-DC .tdms file format"""
from distutils.version import LooseVersion
import pathlib
import time

from .exc import ContourIndexingError  # noqa:F401
from .exc import InvalidTDMSFileFormatError  # noqa:F401
from .exc import IncompleteTDMSFileFormatError  # noqa:F401
from .exc import InvalidVideoFileError  # noqa:F401

try:
    import nptdms
except ModuleNotFoundError:
    NPTDMS_AVAILABLE = False
else:
    if LooseVersion(nptdms.__version__) < LooseVersion("0.23.0"):
        raise ValueError("Please install nptdms>=0.23.0")
    NPTDMS_AVAILABLE = True
    from .event_contour import ContourColumn
    from .event_image import ImageColumn
    from .event_mask import MaskColumn
    from .event_trace import TraceColumn
    from . import naming

import numpy as np

from ... import definitions as dfn
from ...util import hashobj, hashfile

from ..config import Configuration
from ..core import RTDCBase


class RTDC_TDMS(RTDCBase):
    def __init__(self, tdms_path, *args, **kwargs):
        """TDMS file format for RT-DC measurements

        Parameters
        ----------
        tdms_path: str or pathlib.Path
            Path to a '.tdms' measurement file.
        *args:
            Arguments for `RTDCBase`
        **kwargs:
            Keyword arguments for `RTDCBase`

        Attributes
        ----------
        path: pathlib.Path
            Path to the experimental dataset (main .tdms file)
        """
        if not NPTDMS_AVAILABLE:
            raise ModuleNotFoundError(
                "Package `nptdms` required for TDMS format!")
        # Initialize RTDCBase
        super(RTDC_TDMS, self).__init__(*args, **kwargs)

        tdms_path = pathlib.Path(tdms_path)
        # Events is a simple dictionary
        self._events = {}
        self._hash = None
        self.path = tdms_path
        self.title = get_project_name_from_path(tdms_path, append_mx=True)

        # tdms-related convenience properties
        self._fdir = tdms_path.parent
        self._mid = tdms_path.name.split("_")[0]

        self._init_data_with_tdms(tdms_path)

        # Add additional features
        # event images
        self._events["image"] = ImageColumn(self)
        # event contours (requires image)
        self._events["contour"] = ContourColumn(self)
        # event masks (requires contour)
        self._events["mask"] = MaskColumn(self)
        # event traces
        self._events["trace"] = TraceColumn(self)

    def __contains__(self, key):
        ct = False
        if key in ["contour", "image", "mask", "trace"]:
            # Take into account special cases of the tdms file format:
            # tdms features "image", "trace", "contour", and "mask"
            # evaluate to True (len()!=0) if the data exist on disk
            if key in self._events and self._events[key]:
                ct = True
        else:
            ct = super(RTDC_TDMS, self).__contains__(key)
        return ct

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        del self._events["image"]._image_data

    def _init_data_with_tdms(self, tdms_filename):
        """Initializes the current RT-DC dataset with a tdms file.
        """
        tdms_file = nptdms.TdmsFile(str(tdms_filename))
        # time is always there
        table = "Cell Track"
        # Edit naming.dclab2tdms to add features
        for arg in naming.tdms2dclab:
            try:
                data = tdms_file[table][arg].data
            except KeyError:
                pass
            else:
                if data is None or len(data) == 0:
                    # Ignore empty features. npTDMS treats empty
                    # features in the following way:
                    # - in nptdms 0.8.2, `data` is `None`
                    # - in nptdms 0.9.0, `data` is an array of length 0
                    continue
                self._events[naming.tdms2dclab[arg]] = data
        if len(self._events) == 0:
            raise IncompleteTDMSFileFormatError(
                "No usable feature data found in '{}'!".format(tdms_filename))
        # Set up configuration
        config_paths = [self.path.with_name(self._mid + "_para.ini"),
                        self.path.with_name(self._mid + "_camera.ini")]
        for cp in config_paths:
            if not cp.exists():
                raise IncompleteTDMSFileFormatError(
                    "Missing file: {}".format(cp))
        shpin_set = self.path.with_name(self._mid + "_SoftwareSettings.ini")
        if shpin_set.exists():
            config_paths.append(shpin_set)

        tdms_config = Configuration(files=config_paths, disable_checks=True)

        dclab_config = Configuration()

        for cfgii in [naming.configmap, naming.config_map_set]:
            for section in cfgii:
                for pname in cfgii[section]:
                    meta = cfgii[section][pname]
                    convfunc = dfn.get_config_value_func(section, pname)
                    if isinstance(meta, tuple):
                        osec, opar = meta
                        if osec in tdms_config and opar in tdms_config[osec]:
                            val = tdms_config[osec].pop(opar)
                            dclab_config[section][pname] = convfunc(val)
                    else:
                        dclab_config[section][pname] = convfunc(meta)

        # Additional information from log file
        rtfdc_log = self.path.with_name(self._mid + "_log.ini")
        if rtfdc_log.exists():
            with rtfdc_log.open("r", errors="replace") as fd:
                loglines = fd.readlines()
            for line in loglines:
                if line.startswith("[EVENT LOG]"):
                    sv = line.split("]")[1].strip()
                    if sv:
                        dclab_config["setup"]["software version"] = sv

        rtfdc_parm = self.path.with_name("parameters.txt")
        if rtfdc_parm.exists():
            with rtfdc_parm.open("r", errors="replace") as fd:
                parlines = fd.readlines()
            p1 = None
            p2 = None
            p3 = None
            for line in parlines:
                if line.startswith("pulse_led"):
                    fdur = float(line.split()[1])
                    dclab_config["imaging"]["flash duration"] = fdur
                elif line.startswith("numberofchannels"):
                    nc = int(line.split()[1])
                    dclab_config["fluorescence"]["channel count"] = nc
                elif line.startswith("laser488"):
                    p1 = float(line.split()[1])
                    dclab_config["fluorescence"]["laser 1 lambda"] = 488
                    dclab_config["fluorescence"]["laser 1 power"] = p1
                elif line.startswith("laser561"):
                    p2 = float(line.split()[1])
                    dclab_config["fluorescence"]["laser 2 lambda"] = 561
                    dclab_config["fluorescence"]["laser 2 power"] = p2
                elif line.startswith("laser640"):
                    p3 = float(line.split()[1])
                    dclab_config["fluorescence"]["laser 3 lambda"] = 640
                    dclab_config["fluorescence"]["laser 3 power"] = p3
                elif line.startswith("samplerate"):
                    sr = int(float(line.split()[1]))
                    dclab_config["fluorescence"]["sample rate"] = sr
                elif line.startswith("samplesperframe"):
                    spe = int(line.split()[1])
                    dclab_config["fluorescence"]["samples per event"] = spe
                elif line.startswith("Vmin"):
                    vmin = float(line.split()[1])
                    dclab_config["fluorescence"]["signal min"] = vmin
                elif line.startswith("Vmax"):
                    vmax = float(line.split()[1])
                    dclab_config["fluorescence"]["signal max"] = vmax
                elif line.startswith("median_pmt"):
                    mfs = int(line.split()[1])
                    dclab_config["fluorescence"]["trace median"] = mfs
            # Add generic channel names (independent of lasers)
            for ii in range(1, 4):
                chn = "channel {} name".format(ii)
                fln = "fl{}_max".format(ii)
                if fln in self and chn not in dclab_config["fluorescence"]:
                    dclab_config["fluorescence"][chn] = "FL{}".format(ii)
            lc = bool(p1) + bool(p2) + bool(p3)
            dclab_config["fluorescence"]["laser count"] = lc
            li = (p1 is not None) + (p2 is not None) + (p3 is not None)
            dclab_config["fluorescence"]["lasers installed"] = li
            dclab_config["fluorescence"]["channels installed"] = 3

        # Additional information from commented-out log-file (manual)
        with config_paths[0].open("r", errors="replace") as fd:
            lns = [s[1:].strip() for s in fd.readlines() if s.startswith("#")]
            if lns and lns[0] == "[FLUOR]":
                if ("software version" not in dclab_config["setup"]
                        and lns[1].startswith("fRTDC")):
                    dclab_config["setup"]["software version"] = lns[1]
                for ll in lns[2:]:
                    if ("sample rate" not in dclab_config["fluorescence"]
                            and ll.startswith("Samplerate")):
                        val = int(float(ll.split("=")[1]))
                        dclab_config["fluorescence"]["sample rate"] = val
                    elif ("signal min" not in dclab_config["fluorescence"]
                            and ll.startswith("ADCmin")):
                        val = float(ll.split("=")[1])
                        dclab_config["fluorescence"]["signal min"] = val
                    elif ("signal max" not in dclab_config["fluorescence"]
                            and ll.startswith("ADCmax")):
                        val = float(ll.split("=")[1])
                        dclab_config["fluorescence"]["signal max"] = val

        self.config = dclab_config
        self._complete_config_tdms(tdms_config)

        self._init_filters()

        # Load log files
        log_files = config_paths
        for name in [self._mid + "_events.txt",
                     self._mid + "_log.ini",
                     self._mid + "_SoftwareSettings.ini",
                     "FG_Config.mcf",
                     "parameters.txt"]:
            pl = self.path.with_name(name)
            if pl.exists():
                log_files.append(pl)
        for pp in log_files:
            with pp.open("r", errors="replace") as f:
                cfg = [s.strip() for s in f.readlines()]
            self.logs[pp.name] = cfg

    def _complete_config_tdms(self, residual_config={}):
        # remove zero frame rate
        if ("frame rate" in self.config["imaging"]
                and self.config["imaging"]["frame rate"] == 0):
            self.config["imaging"].pop("frame rate")
        # measurement start time
        tse = self.path.stat().st_mtime
        if "time" in self:
            # correct for duration of experiment
            tse -= self["time"][-1]
        loct = time.localtime(tse)
        if "date" not in self.config["experiment"]:
            # Date of measurement ('YYYY-MM-DD')
            datestr = time.strftime("%Y-%m-%d", loct)
            self.config["experiment"]["date"] = datestr
        if "event count" not in self.config["experiment"]:
            # Number of recorded events
            self.config["experiment"]["event count"] = len(self)
        if "sample" not in self.config["experiment"]:
            # Measured sample or user-defined reference
            sample = get_project_name_from_path(self.path)
            self.config["experiment"]["sample"] = sample
        if "time" not in self.config["experiment"]:
            # Start time of measurement ('HH:MM:SS')
            timestr = time.strftime("%H:%M:%S", loct)
            self.config["experiment"]["time"] = timestr
        # fluorescence
        if ("fluorescence" in self.config
            or "fl1_max" in self
            or "fl2_max" in self
                or "fl3_max" in self):
            if "bit depth" not in self.config["fluorescence"]:
                # hardware-defined (always the same)
                self.config["fluorescence"]["bit depth"] = 16
            if "laser 1 power" in self.config["fluorescence"]:
                self.config["fluorescence"]["laser 1 lambda"] = 488.
            if "laser 2 power" in self.config["fluorescence"]:
                self.config["fluorescence"]["laser 2 lambda"] = 561.
            if "laser 3 power" in self.config["fluorescence"]:
                self.config["fluorescence"]["laser 3 lambda"] = 640.
        # fmt_tdms
        if "video frame offset" not in self.config["fmt_tdms"]:
            self.config["fmt_tdms"]["video frame offset"] = 1
        # setup (compatibility to old tdms formats)
        if "flow rate" not in self.config["setup"]:
            self.config["setup"]["flow rate"] = np.nan
        if "channel width" not in self.config["setup"]:
            if "channel width" in residual_config["general"]:
                channel_width = residual_config["general"]["channel width"]
            elif self.config["setup"]["flow rate"] < 0.16:
                channel_width = 20.
            else:
                channel_width = 30.
            self.config["setup"]["channel width"] = channel_width
        # imaging
        if "pixel size" not in self.config["imaging"]:
            self.config["imaging"]["pixel size"] = 0.34
        # medium convention for CellCarrierB
        if ("medium" in self.config["setup"] and
                self.config["setup"]["medium"].lower() == "cellcarrier b"):
            self.config["setup"]["medium"] = "CellCarrierB"
        # replace "+" with ","
        if "module composition" in self.config["setup"]:
            mc = self.config["setup"]["module composition"]
            if mc.count("+"):
                mc2 = ", ".join([m.strip() for m in mc.split("+")])
                self.config["setup"]["module composition"] = mc2

    @staticmethod
    def can_open(h5path):
        """Check whether a given file is in the .tdms file format"""
        return pathlib.Path(h5path).suffix == ".tdms"

    @property
    def hash(self):
        """Hash value based on file name and .ini file content"""
        if self._hash is None:
            # Only hash _camera.ini and _para.ini
            fsh = [self.path.with_name(self._mid + "_camera.ini"),
                   self.path.with_name(self._mid + "_para.ini")]
            tohash = [hashfile(f) for f in fsh]
            tohash.append(self.path.name)
            # Hash a maximum of ~1MB of the tdms file
            tohash.append(hashfile(self.path, blocksize=65536, count=20))
            self._hash = hashobj(tohash)
        return self._hash


def get_project_name_from_path(path, append_mx=False):
    """Get the project name from a path.

    For a path "/home/peter/hans/HLC12398/online/M1_13.tdms" or
    For a path "/home/peter/hans/HLC12398/online/data/M1_13.tdms" or
    without the ".tdms" file, this will return always "HLC12398".

    Parameters
    ----------
    path: str
        path to tdms file
    append_mx: bool
        append measurement number, e.g. "M1"
    """
    path = pathlib.Path(path)
    if path.suffix == ".tdms":
        dirn = path.parent
        mx = path.name.split("_")[0]
    elif path.is_dir():
        dirn = path
        mx = ""
    else:
        dirn = path.parent
        mx = ""

    project = ""
    if mx:
        # check para.ini
        para = dirn / (mx + "_para.ini")
        if para.exists():
            with para.open("r", errors="replace") as fd:
                lines = fd.readlines()
            for line in lines:
                if line.startswith("Sample Name ="):
                    project = line.split("=")[1].strip()
                    break

    if not project:
        # check if the directory contains data or is online
        root1, trail1 = dirn.parent, dirn.name
        root2, trail2 = root1.parent, root1.name
        trail3 = root2.name

        if trail1.lower() in ["online", "offline"]:
            # /home/peter/hans/HLC12398/online/
            project = trail2
        elif (trail1.lower() == "data" and
              trail2.lower() in ["online", "offline"]):
            # this is olis new folder sctructure
            # /home/peter/hans/HLC12398/online/data/
            project = trail3
        else:
            project = trail1

    if append_mx:
        project += " - " + mx

    return project


def get_tdms_files(directory):
    """Recursively find projects based on '.tdms' file endings

    Searches the `directory` recursively and return a sorted list
    of all found '.tdms' project files, except fluorescence
    data trace files which end with `_traces.tdms`.
    """
    path = pathlib.Path(directory).resolve()
    # get all tdms files
    tdmslist = [r for r in path.rglob("*.tdms") if r.is_file()]
    # exclude traces files
    tdmslist = [r for r in tdmslist if not r.name.endswith("_traces.tdms")]
    return sorted(tdmslist)



from distutils.version import LooseVersion
from importlib import import_module
from unittest import mock

import numpy as np
import pytest

import dclab
from dclab.lme4 import Rlme4, bootstrapped_median_distributions, rsetup, rlibs


pytest.importorskip("rpy2")


def standard_datasets(set_region=True):
    features = [
        [100, 99, 80, 120, 140, 150, 100, 100, 110, 111,
         140, 145],  # Larger values (Control Channel1)
        [20, 10, 5, 16, 14, 22, 27, 26, 5, 10, 11, 8, 15, 17,
         20, 9],  # Smaller values (Control Reservoir1)
        [115, 110, 90, 110, 145, 155, 110, 120, 115, 120, 120,
         150, 100, 90, 100],  # Larger values (Control Channel2)
        [30, 30, 15, 26, 24, 32, 37, 36, 15, 20, 21, 18, 25,
         27, 30, 19],  # Smaller values (Control Reservoir2)
        [150, 150, 130, 170, 190, 250, 150, 150, 160, 161, 180, 195, 130,
         120, 125, 130, 125],  # Larger values (Treatment Channel1)
        [2, 1, 5, 6, 4, 2, 7, 6, 5, 10, 1, 8, 5, 7, 2, 9, 11,
         8, 13],  # Smaller values (Treatment Reservoir1)
        [155, 155, 135, 175, 195, 255, 155, 155, 165, 165, 185, 200, 135, 125,
         130, 135, 140, 150, 135, 140],  # Larger values (Treatment Channel2)
        [25, 15, 19, 26, 44, 42, 35, 20, 15, 10, 11, 28, 35, 10, 25,
         13]]  # Smaller values (Treatment Reservoir2)
    regions = ["channel", "reservoir"] * 4
    datasets = []
    for ii, ff in enumerate(features):
        ds = dclab.new_dataset({"deform": ff})
        if set_region:
            region = regions[ii]
        else:
            region = "channel"
        ds.config["setup"]["chip region"] = region
        assert ds.config["setup"]["chip region"] == region
        datasets.append(ds)
    return datasets


def test_differential():
    # Larger values (Channel1)
    a = np.array([100, 99, 80, 120, 140, 150, 100, 100, 110, 111, 140, 145])
    # Smaller values (Reservoir1)
    b = np.array([20, 10, 5, 16, 14, 22, 27, 26, 5, 10, 11, 8, 15, 17, 20, 9])
    result = bootstrapped_median_distributions(a, b, bs_iter=1000)
    assert np.allclose([np.median(result[0])], [110.5])
    assert np.allclose([np.median(result[1])], [14.5])


def test_basic_setup():
    assert rsetup.has_r()
    rsetup.install_lme4()
    assert rsetup.has_lme4()


def test_import_rpy2():
    import rpy2
    from rpy2 import situation
    assert LooseVersion(rpy2.__version__) >= LooseVersion(
        rlibs.RPY2_MIN_VERSION)
    assert situation.get_r_home() is not None


def test_import_submodules_raise_r_unavailable_error():
    import rpy2

    def mocked_import_module(mod):
        if mod == "rpy2.robjects.packages":
            raise rpy2.rinterface_lib.openrlib.ffi.error("Testing")
        else:
            return import_module(mod)

    with mock.patch("importlib.import_module", new=mocked_import_module):
        with pytest.raises(rlibs.ROutdatedError, match="Testing"):
            rlibs.import_r_submodules()


def test_fail_add_same_dataset():
    datasets = standard_datasets(set_region=False)

    rlme4 = Rlme4(model="lmer", feature="deform")
    rlme4.add_dataset(datasets[0], "control", 1)

    with pytest.raises(ValueError, match="has already"):
        rlme4.add_dataset(datasets[1], "control", 1)


def test_fail_too_few_dataset():
    datasets = standard_datasets(set_region=False)

    rlme4 = Rlme4(model="lmer", feature="deform")
    rlme4.add_dataset(datasets[0], "control", 1)
    rlme4.add_dataset(datasets[1], "treatment", 1)

    with pytest.raises(ValueError, match="Linear mixed effects models"
                                         + " require repeated measurements"):
        rlme4.fit()


def test_fail_get_non_existent_data():
    datasets = standard_datasets(set_region=False)

    rlme4 = Rlme4(model="lmer", feature="deform")
    rlme4.add_dataset(datasets[0], "control", 1)
    rlme4.add_dataset(datasets[1], "treatment", 1)
    rlme4.add_dataset(datasets[2], "control", 2)
    rlme4.add_dataset(datasets[3], "treatment", 2)
    rlme4.add_dataset(datasets[3], "treatment", 3)

    with pytest.raises(ValueError, match="Dataset for group 'control'"):
        rlme4.get_feature_data(group="control", repetition=3)


def test_glmer_basic_larger():
    """Original values in a generalized linear mixed model"""
    groups = ['treatment', 'control', 'treatment', 'control',
              'treatment', 'control', 'treatment', 'control']
    repetitions = [1, 1, 2, 2, 3, 3, 4, 4]
    datasets = standard_datasets(set_region=False)

    rlme4 = Rlme4(model="glmer+loglink", feature="deform")
    for ii in range(len(datasets)):
        rlme4.add_dataset(datasets[ii], groups[ii], repetitions[ii])

    res = rlme4.fit()
    assert np.allclose(res["fixed effects intercept"], 15.083832084641697)
    assert np.allclose(res["anova p-value"], 0.00365675950677214)
    assert not rlme4.is_differential()
    assert res["model converged"]


def test_glmer_differential():
    """Differential Deformation in a generalized linear mixed model"""
    groups = ['control', 'control', 'control', 'control', 'treatment',
              'treatment', 'treatment', 'treatment']
    repetitions = [1, 1, 2, 2, 1, 1, 2, 2]
    datasets = standard_datasets()
    rlme4 = Rlme4(model="glmer+loglink", feature="deform")
    for ii in range(len(datasets)):
        rlme4.add_dataset(datasets[ii], groups[ii], repetitions[ii])

    res = rlme4.fit()
    assert np.allclose(res["fixed effects intercept"], 93.55789706339498)
    assert rlme4.is_differential()
    assert rlme4.model == "glmer+loglink"
    assert np.allclose(res["anova p-value"], 0.000556063024310929)
    assert res["model converged"]


def test_lmer_basic():
    groups = ['control', 'treatment', 'control', 'treatment']
    repetitions = [1, 1, 2, 2]
    features = [
        [100, 99, 80, 120, 140, 150, 100, 100, 110, 111, 140, 145],
        [115, 110, 90, 110, 145, 155, 110, 120, 115, 120, 120, 150,
         100, 90, 100],
        [150, 150, 130, 170, 190, 250, 150, 150, 160,
         161, 180, 195, 130, 120, 125, 130, 125],
        [155, 155, 135, 175, 195, 255, 155, 155, 165, 165,
         185, 200, 135, 125, 130, 135, 140, 150, 135, 140]
    ]
    datasets = []
    for ff in features:
        ds = dclab.new_dataset({"deform": ff})
        ds.config["setup"]["chip region"] = "channel"
        assert ds.config["setup"]["chip region"] == "channel"
        datasets.append(ds)

    rlme4 = Rlme4(model="lmer", feature="deform")
    for ii in range(len(datasets)):
        rlme4.add_dataset(datasets[ii], groups[ii], repetitions[ii])

    res = rlme4.fit()
    assert np.allclose(res["anova p-value"], 0.8434625179432129)
    assert np.allclose(res["fixed effects intercept"], 136.6365047987075)
    assert np.allclose(res["fixed effects treatment"], 1.4085644911191584,
                       atol=0, rtol=1e-3)
    assert np.allclose(res["fixed effects intercept"],
                       np.mean(res["fixed effects repetitions"], axis=1)[0])
    assert np.allclose(res["fixed effects treatment"],
                       np.mean(res["fixed effects repetitions"], axis=1)[1])
    assert not res["is differential"]
    assert res["feature"] == "deform"
    assert res["model"] == "lmer"
    assert res["model converged"]


def test_lmer_basic_filtering():
    groups = ['control', 'treatment', 'control', 'treatment']
    repetitions = [1, 1, 2, 2]
    features = [
        [100, 99, 80, 120, 140, 150, 100, 100, 110, 111, 140, 145],
        [115, 110, 90, 110, 145, 155, 110, 120, 115, 120, 120, 150,
         100, 90, 100],
        [150, 150, 130, 170, 190, 250, 150, 150, 160,
         161, 180, 195, 130, 120, 125, 130, 125],
        [155, 155, 135, 175, 195, 255, 155, 155, 165, 165,
         185, 200, 135, 125, 130, 135, 140, 150, 135, 140]
    ]
    datasets = []
    for ff in features:
        ds = dclab.new_dataset({"deform": ff})
        ds.config["setup"]["chip region"] = "channel"
        # apply some filters
        datasets.append(ds)

    rlme4 = Rlme4(model="lmer", feature="deform")
    for ii in range(len(datasets)):
        rlme4.add_dataset(datasets[ii], groups[ii], repetitions[ii])

    res = rlme4.fit()
    assert np.allclose(res["anova p-value"], 0.8434625179432129)

    # filters should have an effect

    for ds in datasets:
        ds.filter.manual[:4] = False
        ds.apply_filter()

    res2 = rlme4.fit()
    assert not np.allclose(res["anova p-value"], res2["anova p-value"])


def test_lmer_basic_larger():
    """Original values in a linear mixed model

    'Reservoir' measurements are now Controls and 'Channel'
    measurements are Treatments. This does not use differential
    deformation.
    """
    groups = ['treatment', 'control', 'treatment', 'control',
              'treatment', 'control', 'treatment', 'control']
    repetitions = [1, 1, 2, 2, 3, 3, 4, 4]
    datasets = standard_datasets(set_region=False)

    rlme4 = Rlme4(model="lmer", feature="deform")
    for ii in range(len(datasets)):
        rlme4.add_dataset(datasets[ii], groups[ii], repetitions[ii])

    res = rlme4.fit()
    assert np.allclose(res["fixed effects intercept"], 17.171341507432501)
    assert np.allclose(res["anova p-value"], 0.000331343267412872)
    assert not rlme4.is_differential()
    assert res["model converged"]


def test_lmer_basic_nan():
    groups = ['control', 'treatment', 'control', 'treatment']
    repetitions = [1, 1, 2, 2]
    features = [
        [100, np.inf, 80, np.nan, 140, 150, 100, 100, 110, 111, 140, 145],
        [115, 110, 90, 110, 145, 155, 110, 120, 115, 120, 120, 150,
         100, 90, 100],
        [150, 150, 130, 170, 190, 250, 150, 150, 160,
         161, 180, 195, 130, 120, 125, 130, 125],
        [155, 155, 135, 175, 195, 255, 155, 155, 165, 165,
         185, 200, 135, 125, 130, 135, 140, 150, 135, 140]
    ]
    datasets = []
    for ff in features:
        ds = dclab.new_dataset({"deform": ff})
        ds.config["setup"]["chip region"] = "channel"
        assert ds.config["setup"]["chip region"] == "channel"
        datasets.append(ds)

    rlme4 = Rlme4(model="lmer", feature="deform")
    for ii in range(len(datasets)):
        rlme4.add_dataset(datasets[ii], groups[ii], repetitions[ii])

    res = rlme4.fit()
    assert np.allclose(res["fixed effects intercept"], 137.37179302516199186)
    # assert res["model converged"]  # does not converge on Travis-CI


def test_lmer_differential():
    """Differential Deformation in a linear mixed model"""
    groups = ['control', 'control', 'control', 'control', 'treatment',
              'treatment', 'treatment', 'treatment']
    repetitions = [1, 1, 2, 2, 1, 1, 2, 2]
    datasets = standard_datasets()
    rlme4 = Rlme4(model="lmer", feature="deform")
    for ii in range(len(datasets)):
        rlme4.add_dataset(datasets[ii], groups[ii], repetitions[ii])

    res = rlme4.fit()
    assert np.allclose(res["fixed effects intercept"], 93.693750004463098)
    assert rlme4.is_differential()
    assert rlme4.model == "lmer"
    assert np.allclose(res["anova p-value"], 0.000602622503360039)
    assert np.allclose(res["fixed effects intercept"],
                       np.mean(res["fixed effects repetitions"], axis=1)[0])
    assert np.allclose(res["fixed effects treatment"],
                       np.mean(res["fixed effects repetitions"], axis=1)[1])
    assert res["model converged"]


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""ML: Using RT-DC data with tensorflow

We use tensorflow to distinguish between beads and cells using
scalar features only. The example data is taken from a `reference
dataset on DCOR <https://dcor.mpl.mpg.de/dataset/figshare-7771184-v2>`_.
The classification accuracy using only the inputs ``area_ratio``,
``area_um``, ``bright_sd``, and ``deform`` reaches values above 95%.

.. warning::

   This example neglects a lot of important aspects of machine learning
   with RT-DC data (e.g. brightness normalization) and it is a very
   easy task (beads are smaller than cells). Thus, this example should
   only be considered as a technical guide on how tensorflow can be used
   with RT-DC data.

.. note::

   What happens when you add ``"bright_avg"`` to the ``features`` list?
   Can you explain the result?

Apparently, debris in the cell dataset is classified as beads.
We could have gotten around that by filtering the input data before
inference. In addition, some beads get classified as cells as well.
This is a result of the limited features used for training/inference.
Under normal cirumstances, you would investigate other features in
order to improve the model prediction.
"""
import matplotlib.pylab as plt
import numpy as np
import tensorflow as tf
from dclab.rtdc_dataset.feat_anc_ml import hook_tensorflow

tf.random.set_seed(42)  # for reproducibility

# https://dcor.mpl.mpg.de/dataset/figshare-7771184-v2
dcor_ids = ["fb719fb2-bd9f-817a-7d70-f4002af916f0",
            "f7fa778f-6abd-1b53-ae5f-9ce12601d6f8"]
labels = [0, 1]  # 0: beads, 1: cells
features = ["area_ratio", "area_um", "bright_sd", "deform"]

# obtain train and test datasets
train, test = hook_tensorflow.assemble_tf_dataset_scalars(
    dc_data=dcor_ids,  # can also be list of paths or datasets
    labels=labels,
    feature_inputs=features,
    split=.8)

# build the model
model = tf.keras.Sequential(
    layers=[
        tf.keras.layers.Input(shape=(len(features),)),
        tf.keras.layers.Dense(128, activation='relu'),
        tf.keras.layers.Dense(32),
        tf.keras.layers.Dropout(0.2),
        tf.keras.layers.Dense(2)
    ],
    name="scalar_features"
)

# fit the model to the training data
loss_fn = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)
model.compile(optimizer='adam', loss=loss_fn, metrics=['accuracy'])
model.fit(train, epochs=5)

# show accuracy using test data (loss: 0.1139 - accuracy: 0.9659)
model.evaluate(test, verbose=2)

# predict classes of the test data
probability_model = tf.keras.Sequential([model, tf.keras.layers.Softmax()])
y_test = np.concatenate([y for x, y in test], axis=0)
predict = np.argmax(probability_model.predict(test), axis=1)

# take a few exemplary events from true and false classification
false_cl = np.where(predict != y_test)[0]
true_cl = np.where(predict == y_test)[0]
num_events = min(4, min(len(true_cl), len(false_cl)))

false_images = hook_tensorflow.get_dataset_event_feature(
    dc_data=dcor_ids,
    feature="image",
    tf_dataset_indices=false_cl[:num_events],
    split_index=1,
    split=.8)

true_images = hook_tensorflow.get_dataset_event_feature(
    dc_data=dcor_ids,
    feature="image",
    tf_dataset_indices=true_cl[:num_events],
    split_index=1,
    split=.8)

fig = plt.figure(figsize=(8, 7))

for ii in range(num_events):
    title_true = ("cell" if y_test[true_cl[[ii]]] else "bead") + " (correct)"
    title_false = ("cell" if predict[false_cl[ii]] else "bead") + " (wrong)"
    ax1 = plt.subplot(num_events, 2, 2*ii+1, title=title_true)
    ax2 = plt.subplot(num_events, 2, 2*(ii + 1), title=title_false)
    ax1.axis("off")
    ax2.axis("off")
    ax1.imshow(true_images[ii], cmap="gray")
    ax2.imshow(false_images[ii], cmap="gray")

plt.tight_layout()
plt.show()



# flake8: noqa: F401
import warnings

from .feat_const import (
    FEATURES_SCALAR, FEATURES_NON_SCALAR, FLUOR_TRACES,
    # these should not be used
    feature_names, feature_labels, feature_name2label,
    # this one should also not be used, but we wait with deprecation,
    # because Shape-Out heavily relies on it (it shouldn't)
    scalar_feature_names
    )
from .feat_logic import (
    check_feature_shape, feature_exists, get_feature_label,
    scalar_feature_exists
    )
from .meta_const import (
    CFG_ANALYSIS, CFG_METADATA, config_keys,
    # these should not be used (in contrast, `config_keys` is quite
    # important for iterating over all internal configuration keys)
    config_descr, config_funcs, config_types,
    )
from .meta_logic import (
    config_key_exists,
    get_config_value_descr, get_config_value_func, get_config_value_type
    )


class DeprecatedObject:
    def __init__(self, anobject, name, new):
        self.anobject = anobject
        self.message = f"dclab.dfn.{name} is deprecated, please use " \
            + f"dclab.dfn.{new} instead."

    def __getattr__(self, item):
        warnings.warn(self.message, DeprecationWarning)
        return self.anobject.__getattr__(item)

    def __getitem__(self, item):
        warnings.warn(self.message, DeprecationWarning)
        return self.anobject.__getitem__(item)

    def __iter__(self):
        warnings.warn(self.message, DeprecationWarning)
        return self.anobject.__iter__()

    def __contains__(self, item):
        warnings.warn(self.message, DeprecationWarning)
        return self.anobject.__contains__(item)


config_descr = DeprecatedObject(config_descr,
                                "config_descr",
                                "get_config_value_descr",
                                )

config_funcs = DeprecatedObject(config_funcs,
                                "config_funcs",
                                "get_config_value_func",
                                )

config_types = DeprecatedObject(config_types,
                                "config_types",
                                "get_config_value_type",
                                )

feature_names = DeprecatedObject(feature_names,
                                 "feature_names",
                                 "feature_exists",
                                 )

feature_labels = DeprecatedObject(feature_labels,
                                  "feature_labels",
                                  "get_feature_label",
                                  )

feature_name2label = DeprecatedObject(feature_name2label,
                                      "feature_name2label",
                                      "get_feature_label",
                                      )



import numbers

import numpy as np


def f2dfloatarray(value):
    """numpy floating point array"""
    return np.array(value, dtype=float)


def fbool(value):
    """boolean"""
    if isinstance(value, str):
        value = value.lower()
        if value == "false":
            value = False
        elif value == "true":
            value = True
        elif value:
            value = bool(float(value))
        else:
            raise ValueError("Empty string provided for fbool!")
    else:
        value = bool(float(value))
    return value


def fint(value):
    """integer"""
    if isinstance(value, str):
        # strings might have been saved wrongly as booleans
        value = value.lower()
        if value == "false":
            value = 0
        elif value == "true":
            value = 1
        elif value:
            value = int(float(value))
        else:
            raise ValueError("Empty string provided for fint!")
    else:
        value = int(float(value))
    return value


def fintlist(alist):
    """A list of integers"""
    outlist = []
    if not isinstance(alist, (list, tuple)):
        # we have a string (comma-separated integers)
        alist = alist.strip().strip("[] ").split(",")
    for it in alist:
        if it:
            outlist.append(fint(it))
    return outlist


def lcstr(astr):
    """lower-case string"""
    return astr.lower()


#: maps functions to their expected output types
func_types = {
    f2dfloatarray: np.ndarray,
    fbool: (bool, np.bool_),
    fint: numbers.Integral,
    fintlist: list,
    float: numbers.Number,
    lcstr: str}



import pytest

from dclab import new_dataset


from helper_methods import example_data_dict, retrieve_data


def test_hash_dict():
    ddict = example_data_dict()
    ds = new_dataset(ddict)
    assert ds.hash == "5bd08693349fd40369860474e3dab144"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hash_hierarchy():
    pytest.importorskip("nptdms")
    tdms_path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    ds1 = new_dataset(tdms_path)
    ds2 = new_dataset(ds1)
    assert ds2.hash == "3e942ba6e1cb333d3607edaba5f2c618"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hash_tdms():
    pytest.importorskip("nptdms")
    tdms_path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    ds = new_dataset(tdms_path)
    assert ds.hash == "92601489292dc9bf9fc040f87d9169c0"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()




#: List of scalar (one scalar value per event) features. This
#: list does not include the `ml_score_???` features. If you
#: need find out whether a feature name is valid, please use
#: :func:`is_valid_feature`.
FEATURES_SCALAR = [
    ["area_cvx", "Convex area [px]"],
    # area_msd is the contour moment M00
    ["area_msd", "Measured area [px]"],
    ["area_ratio", "Porosity (convex to measured area ratio)"],
    # area_um is the convex area per definition
    ["area_um", "Area [µm²]"],
    ["aspect", "Aspect ratio of bounding box"],
    ["bright_avg", "Brightness average within contour [a.u.]"],
    ["bright_sd", "Brightness SD within contour [a.u.]"],
    ["circ", "Circularity"],
    # deform is computed from the convex contour
    ["deform", "Deformation"],
    ["emodulus", "Young's modulus [kPa]"],
    # fl*_area, fl*_pos, and fl*_width values correspond to the
    # object for which the contour was found. For high concentrations,
    # these values could be error-prone due to the assignment from
    # false objects.
    ["fl1_area", "FL-1 area of peak [a.u.]"],
    # fl1_dist is set to zero if there is only one peak
    ["fl1_dist", "FL-1 distance between two first peaks [µs]"],
    ["fl1_max", "FL-1 maximum [a.u.]"],
    ["fl1_max_ctc", "FL-1 maximum, crosstalk-corrected [a.u.]"],
    ["fl1_npeaks", "FL-1 number of peaks"],
    ["fl1_pos", "FL-1 position of peak [µs]"],
    ["fl1_width", "FL-1 width [µs]"],
    ["fl2_area", "FL-2 area of peak [a.u.]"],
    ["fl2_dist", "FL-2 distance between two first peaks [µs]"],
    ["fl2_max", "FL-2 maximum [a.u.]"],
    ["fl2_max_ctc", "FL-2 maximum, crosstalk-corrected [a.u.]"],
    ["fl2_npeaks", "FL-2 number of peaks"],
    ["fl2_pos", "FL-2 position of peak [µs]"],
    ["fl2_width", "FL-2 width [µs]"],
    ["fl3_area", "FL-3 area of peak [a.u.]"],
    ["fl3_dist", "FL-3 distance between two first peaks [µs]"],
    ["fl3_max", "FL-3 maximum [a.u.]"],
    ["fl3_max_ctc", "FL-3 maximum, crosstalk-corrected [a.u.]"],
    ["fl3_npeaks", "FL-3 number of peaks"],
    ["fl3_pos", "FL-3 position of peak [µs]"],
    ["fl3_width", "FL-3 width [µs]"],
    ["frame", "Video frame number"],
    ["g_force", "Gravitational force in multiples of g"],
    # index starts with 1
    ["index", "Event index (Dataset)"],
    # index_online may have missing values (#71)
    ["index_online", "Event index (Online)"],
    # The inertia ratios of the event contours are defined by the
    # central second order moments of area (sqrt(m20/m02).
    ["inert_ratio_cvx", "Inertia ratio of convex contour"],
    ["inert_ratio_prnc", "Principal inertia ratio of raw contour"],
    ["inert_ratio_raw", "Inertia ratio of raw contour"],
    # This is an ancillary integer feature for visualizing the class
    # membership of individual events based on the `ml_score_???`
    # features.
    ["ml_class", "Most probable ML class"],
    ["nevents", "Total number of events in the same image"],
    ["pc1", "Principal component 1"],
    ["pc2", "Principal component 2"],
    # pos_x and pos_y are computed from the contour moments
    # "m10"/"m00" and "m01"/"m00" of the convex hull of "contour"
    ["pos_x", "Position along channel axis [µm]"],
    ["pos_y", "Position lateral in channel [µm]"],
    ["size_x", "Bounding box size x [µm]"],
    ["size_y", "Bounding box size y [µm]"],
    ["temp", "Chip temperature [°C]"],
    ["temp_amb", "Ambient temperature [°C]"],
    ["tilt", "Absolute tilt of raw contour"],
    ["time", "Event time [s]"],
    # Volume is computed from the raw contour (i.e. with exclusions).
    # Fun fact: If we had decided to compute it from the convex contour,
    # then we would have close to none pixelation effects ¯\_(ツ)_/¯.
    ["volume", "Volume [µm³]"],
]
# Add userdef features
for _i in range(10):
    FEATURES_SCALAR.append(["userdef{}".format(_i),
                            "User defined {}".format(_i)
                            ])

#: list of non-scalar features
FEATURES_NON_SCALAR = [
    # This is a (M, 2)-shaped array with integer contour coordinates
    ["contour", "Event contour"],
    ["image", "Gray scale event image"],
    ["image_bg", "Gray scale event background image"],
    # This is the contour with holes filled
    ["mask", "Binary mask labeling the event in the image"],
    # See FLUOR_TRACES for valid keys
    ["trace", "Dictionary of fluorescence traces"],
]

#: List of fluorescence traces
FLUOR_TRACES = [
    "fl1_median",
    "fl1_raw",
    "fl2_median",
    "fl2_raw",
    "fl3_median",
    "fl3_raw",
]


# FEATURE convenience lists and dicts

#: list of feature names
feature_names = [_cc[0] for _cc in FEATURES_SCALAR + FEATURES_NON_SCALAR]

#: list of feature labels (same order as :const:`feature_names`
feature_labels = [_cc[1] for _cc in FEATURES_SCALAR + FEATURES_NON_SCALAR]

#: dict for converting feature names to labels
feature_name2label = {}

for _cc in FEATURES_SCALAR + FEATURES_NON_SCALAR:
    feature_name2label[_cc[0]] = _cc[1]

#: list of scalar feature names
scalar_feature_names = [_cc[0] for _cc in FEATURES_SCALAR]



import shutil
import tempfile
import time

TMPDIR = tempfile.mkdtemp(prefix=time.strftime(
    "dclab_test_%H.%M_"))


def pytest_configure(config):
    """
    Allows plugins and conftest files to perform initial configuration.
    This hook is called for every plugin and initial conftest
    file after command line options have been parsed.
    """
    tempfile.tempdir = TMPDIR


def pytest_unconfigure(config):
    """
    called before test process is exited.
    """
    shutil.rmtree(TMPDIR, ignore_errors=True)



import numpy as np
import pytest

import dclab

from helper_methods import example_data_dict


def test_stat_simple():
    ddict = example_data_dict(size=5085, keys=["area_um", "tilt"])
    ds = dclab.new_dataset(ddict)

    head, vals = dclab.statistics.get_statistics(ds, features=["tilt"])

    for h, v in zip(head, vals):
        if h.lower() == "flow rate":
            assert np.isnan(v)  # backwards compatibility!
        elif h.lower() == "events":
            assert v == 5085
        elif h.lower() == "%-gated":
            assert v == 100
        elif h.lower().startswith("sd "):
            assert np.allclose(v, 0.288990352083)
        elif h.lower().startswith("median "):
            assert np.allclose(v, 0.494188566376)
        elif h.lower().startswith("mode "):
            assert np.allclose(v, 0.260923009639)
        elif h.lower().startswith("mean "):
            assert np.allclose(v, 0.497743857424)


def test_stat_occur():
    ddict = example_data_dict(size=5085, keys=["area_um", "deform"])
    ds = dclab.new_dataset(ddict)

    head1, vals1 = dclab.statistics.get_statistics(ds, features=["deform"])
    head2, vals2 = dclab.statistics.get_statistics(
        ds, methods=["Events", "Mean"])
    headf, valsf = dclab.statistics.get_statistics(ds)

    # disable filtering (there are none anyway) to cover a couple more lines:
    ds.config["filtering"]["enable filters"] = False
    headn, valsn = dclab.statistics.get_statistics(ds)

    for item in zip(head1, vals1):
        assert item in zip(headf, valsf)

    for item in zip(head2, vals2):
        assert item in zip(headf, valsf)

    for item in zip(headn, valsn):
        assert item in zip(headf, valsf)


def test_flow_rate():
    ddict = example_data_dict(size=77, keys=["area_um", "deform"])
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.172

    head1, vals1 = dclab.statistics.get_statistics(ds, features=["deform"])
    head2, vals2 = dclab.statistics.get_statistics(
        ds, methods=["Events", "Mean"])
    headf, valsf = dclab.statistics.get_statistics(ds)

    # disable filtering (there are none anyway) to cover a couple more lines:
    ds.config["filtering"]["enable filters"] = False
    headn, valsn = dclab.statistics.get_statistics(ds)

    for item in zip(head1, vals1):
        assert item in zip(headf, valsf)

    for item in zip(head2, vals2):
        assert item in zip(headf, valsf)

    for item in zip(headn, valsn):
        assert item in zip(headf, valsf)


@pytest.mark.filterwarnings('ignore::dclab.statistics.BadMethodWarning')
def test_false_method():
    def bad_method(x):
        return x + 1
    dclab.statistics.Statistics(name="bad",
                                req_feature=False,
                                method=bad_method)
    ddict = example_data_dict(size=77, keys=["area_um", "deform"])
    ds = dclab.new_dataset(ddict)
    head1, vals1 = dclab.statistics.get_statistics(ds, features=["deform"])
    out = {}
    for h, v in zip(head1, vals1):
        out[h] = v
    assert np.isnan(out["bad"])

    # clean up
    mth = dclab.statistics.Statistics.available_methods
    for k in mth:
        if k == "bad":
            mth.pop(k)
            break


def test_stats_without_features():
    ddict = example_data_dict(size=5085, keys=["area_um", "deform"])
    ds = dclab.new_dataset(ddict)

    head1, vals1 = dclab.statistics.get_statistics(ds, features=["deform"])
    head2, vals2 = dclab.statistics.get_statistics(
        ds, methods=["Events", "Mean"], features=[])
    head3, vals2 = dclab.statistics.get_statistics(
        ds, methods=[], features=[])

    assert "Events" in head1
    assert "Events" in head2
    assert "Events" not in head3



"""
.. versionadded:: 0.33.0
"""
from ..definitions import feat_logic

from .fmt_hierarchy import RTDC_Hierarchy


_registered_temporary_features = []


def deregister_all():
    """Deregisters all temporary features"""
    for feat in list(_registered_temporary_features):
        deregister_temporary_feature(feat)


def deregister_temporary_feature(feature):
    """Convenience function for deregistering a temporary feature

    This method is mostly used during testing. It does not
    remove the actual feature data from any dataset; the data
    will stay in memory but is not accessible anymore through
    the public methods of the :class:`RTDCBase` user interface.
    """
    if feature in _registered_temporary_features:
        _registered_temporary_features.remove(feature)
        feat_logic.feature_deregister(feature)


def register_temporary_feature(feature, label=None, is_scalar=True):
    """Register a new temporary feature

    Temporary features are custom features that can be defined ad hoc
    by the user. Temporary features are helpful when the integral
    features are not enough, e.g. for prototyping, testing, or
    collating with other data. Temporary features allow you to
    leverage the full functionality of :class:`RTDCBase` with
    your custom features (no need to go for a custom `pandas.Dataframe`).

    Parameters
    ----------
    feature: str
        Feature name; allowed characters are lower-case letters,
        digits, and underscores
    label: str
        Feature label used e.g. for plotting
    is_scalar: bool
        Whether or not the feature is a scalar feature
    """
    feat_logic.feature_register(feature, label, is_scalar)
    _registered_temporary_features.append(feature)


def set_temporary_feature(rtdc_ds, feature, data):
    """Set temporary feature data for a dataset

    Parameters
    ----------
    rtdc_ds: dclab.RTDCBase
        Dataset for which to set the feature. Note that temporary
        features cannot be set for hierarchy children and that the
        length of the feature `data` must match the number of events
        in `rtdc_ds`.
    feature: str
        Feature name
    data: np.ndarray
        The data
    """
    if not feat_logic.feature_exists(feature):
        raise ValueError(
            "Temporary feature '{}' has not been registered!".format(feature))
    if isinstance(rtdc_ds, RTDC_Hierarchy):
        raise NotImplementedError("Setting temporary features for hierarchy "
                                  "children not implemented yet!")
    if len(data) != len(rtdc_ds):
        raise ValueError("The temporary feature `data` must have same length "
                         "as the dataset. Expected length {}, got length "
                         "{}!".format(len(rtdc_ds), len(data)))
    feat_logic.check_feature_shape(feature, data)
    rtdc_ds._usertemp[feature] = data



import h5py
import numpy as np
import pytest

import dclab
from dclab.rtdc_dataset.feat_temp import deregister_all
from dclab.rtdc_dataset.fmt_hierarchy import ChildNDArray

from helper_methods import retrieve_data


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    pass
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    # Remove all temporary features
    deregister_all()


def test_bad_feature_name():
    """Basic test of a temporary feature"""
    with pytest.raises(ValueError):
        dclab.register_temporary_feature("Peter-Pan")


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_basic():
    """Basic test of a temporary feature"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        dclab.register_temporary_feature("my_special_feature")
        dclab.set_temporary_feature(rtdc_ds=ds,
                                    feature="my_special_feature",
                                    data=np.arange(len(ds)))
        assert ds["my_special_feature"][0] == 0


def test_basic_feature_exists():
    """Basic test of a temporary feature"""
    dclab.register_temporary_feature("my_special_feature")
    assert dclab.dfn.feature_exists("my_special_feature")


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_export_and_load():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    # register temporary feature
    dclab.register_temporary_feature(feature="fl1_mean")

    with dclab.new_dataset(h5path) as ds:
        # extract the feature information from the dataset
        fl1_mean = np.array(
            [np.mean(ds["trace"]["fl1_raw"][ii]) for ii in range(len(ds))])
        # set the data
        dclab.set_temporary_feature(rtdc_ds=ds, feature="fl1_mean",
                                    data=fl1_mean)
        # export the data to a new file
        expath = h5path.with_name("exported.rtdc")
        ds.export.hdf5(expath, features=ds.features_innate + ["fl1_mean"])

    # make sure that worked
    with h5py.File(expath, "r") as h5:
        assert "fl1_mean" in h5["events"]
        assert np.allclose(h5["events"]["fl1_mean"], fl1_mean)

    # now check again with dclab
    with dclab.new_dataset(expath) as ds2:
        assert "fl1_mean" in ds2
        assert np.allclose(ds2["fl1_mean"], fl1_mean)

        # and a control check
        deregister_all()
        assert "fl1_mean" not in ds2


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_filtering():
    """Filtering with features, same example as in docs"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        dclab.register_temporary_feature(feature="fl1_mean")
        fl1_mean = np.array(
            [np.mean(ds["trace"]["fl1_raw"][ii]) for ii in range(len(ds))])
        dclab.set_temporary_feature(rtdc_ds=ds,
                                    feature="fl1_mean",
                                    data=fl1_mean)
        ds.config["filtering"]["fl1_mean min"] = 4
        ds.config["filtering"]["fl1_mean max"] = 200
        ds.apply_filter()
        assert np.sum(ds.filter.all) == 1
        assert ds.filter.all[1]


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hierarchy_not_supported():
    """Test for RTDCHierarchy (does not work)"""
    # Hi there,
    # if you are here. this means that this test failed and
    # you just implemented temporary features for hierarchy
    # datasets. I'm fine with that, just make sure that the
    # root parent gets nan values.
    # Cheers,
    # Paul
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        child = dclab.new_dataset(ds)
        dclab.register_temporary_feature("my_special_feature")
        with pytest.raises(NotImplementedError):
            dclab.set_temporary_feature(rtdc_ds=child,
                                        feature="my_special_feature",
                                        data=np.arange(len(child)))


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_inherited_non_scalar():
    """Accessing inherited non-innate, non-scalar features"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        dclab.register_temporary_feature(feature="image_copy", is_scalar=False)
        dclab.set_temporary_feature(rtdc_ds=ds, feature="image_copy",
                                    data=ds["image"][:])
        ds.filter.manual[2] = False
        ch = dclab.new_dataset(ds)
        assert isinstance(ch["image_copy"], ChildNDArray)
        assert ch["image_copy"][:].ndim == 3
        assert isinstance(ch["image_copy"][:], np.ndarray)
        assert np.all(ch["image_copy"][2] == ds["image_copy"][3])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_inherited_scalar():
    """Accessing inherited scalar feature should return np.ndarray"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        dclab.register_temporary_feature("my_special_feature")
        dclab.set_temporary_feature(rtdc_ds=ds,
                                    feature="my_special_feature",
                                    data=np.arange(len(ds)))
        ds.filter.manual[2] = False
        ch = dclab.new_dataset(ds)
        assert "my_special_feature" in ch
        assert isinstance(ch["my_special_feature"], np.ndarray)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_load_temporary_feature_from_disk():
    """Load a temporary feature from a file on disk"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        fl1_mean = np.array(
            [np.mean(ds["trace"]["fl1_raw"][ii]) for ii in range(len(ds))])
    with h5py.File(h5path, "a") as h5:
        h5["events"]["fl1_mean"] = fl1_mean
    # make sure it does not work without registration
    with dclab.new_dataset(h5path) as ds:
        assert "fl1_mean" not in ds, "not registered yet"
    # now, register
    dclab.register_temporary_feature(feature="fl1_mean")
    with dclab.new_dataset(h5path) as ds:
        assert "fl1_mean" in ds
        assert np.all(fl1_mean == ds["fl1_mean"])
        assert "fl1_mean" in ds._events, "registered feature loaded as usual"
        assert "fl1_mean" not in ds._usertemp, "because it is in the file"
        # make sure the feature is not available anymore when deregistered
        deregister_all()
        assert "fl1_mean" not in ds, "deregistered features are hidden..."
        assert "fl1_mean" in ds._events._features, "..but are still there"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_register_after_loading():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        fl1_mean = np.array(
            [np.mean(ds["trace"]["fl1_raw"][ii]) for ii in range(len(ds))])
    with h5py.File(h5path, "a") as h5:
        h5["events"]["fl1_mean"] = fl1_mean
    with dclab.new_dataset(h5path) as ds:
        assert "fl1_mean" not in ds
        dclab.register_temporary_feature(feature="fl1_mean")
        assert "fl1_mean" in ds


def test_try_existing_feature_fails():
    """Basic test of a temporary feature"""
    with pytest.raises(ValueError):
        dclab.register_temporary_feature("deform")


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_with_user_config_section():
    """Use a temporary feature with the user defined config section"""
    # add some metadata to the user config section
    metadata = {"channel": True,
                "n_constrictions": 3}
    ds = dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    ds.config["user"].update(metadata)
    assert ds.config["user"] == metadata
    area_of_region = ds["area_um"] * ds.config["user"]["n_constrictions"]

    dclab.register_temporary_feature("area_of_region")
    dclab.set_temporary_feature(rtdc_ds=ds,
                                feature="area_of_region",
                                data=area_of_region)
    area_of_region1 = ds["area_of_region"]
    area_of_region1_calc = (ds["area_um"] *
                            ds.config["user"]["n_constrictions"])
    assert np.allclose(area_of_region1, area_of_region1_calc)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_wrong_data_shape_1():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        dclab.register_temporary_feature("my_special_feature", is_scalar=False)
        with pytest.raises(ValueError):
            dclab.set_temporary_feature(rtdc_ds=ds,
                                        feature="my_special_feature",
                                        data=np.arange(len(ds)))


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_wrong_data_shape_2():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        dclab.register_temporary_feature("my_special_feature", is_scalar=True)
        data = np.arange(len(ds)*2).reshape(-1, 2)
        with pytest.raises(ValueError):
            dclab.set_temporary_feature(rtdc_ds=ds,
                                        feature="my_special_feature",
                                        data=data)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_wrong_length():
    """temporary feature should have same length"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        dclab.register_temporary_feature("my_special_feature")
        with pytest.raises(ValueError):
            dclab.set_temporary_feature(rtdc_ds=ds,
                                        feature="my_special_feature",
                                        data=np.arange(len(ds)//2))


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_wrong_name():
    """temporary feature should have same length"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        dclab.register_temporary_feature("my_special_feature")
        with pytest.raises(ValueError):
            dclab.set_temporary_feature(rtdc_ds=ds,
                                        feature="my_other_feature",
                                        data=np.arange(len(ds)))


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()
            deregister_all()



from .ancillary_feature import AncillaryFeature  # noqa: F401
from . import af_basic
from . import af_emodulus
from . import af_fl_max_ctc
from . import af_image_contour
from . import af_ml_class


#: features whose computation is fast
FEATURES_RAPID = [
    "area_ratio",
    "area_um",
    "aspect",
    "deform",
    "index",
    "time",
]


af_basic.register()
af_emodulus.register()
af_fl_max_ctc.register()
af_image_contour.register()
af_ml_class.register()



import abc

import numpy as np


class BaseModel(abc.ABC):
    def __init__(self, bare_model, inputs, outputs, info=None):
        """
        Parameters
        ----------
        bare_model:
            Underlying ML model
        inputs: list of str
            List of model input features, e.g.
            ``["deform", "area_um"]``
        outputs: list of str
            List of output features the model provides in that order, e.g.
            ``["ml_score_rbc", "ml_score_rt1", "ml_score_tfe"]``
        info: dict
            Dictionary with model metadata
        """
        if info is None:
            info = {}
        info.setdefault("input features", inputs)
        info.setdefault("output features", outputs)
        self.bare_model = bare_model
        self.inputs = inputs
        self.outputs = outputs
        self.info = info
        self.output_labels = info.get("output labels") or outputs

    @staticmethod
    def all_formats():
        """Dict of dictionaries containing all model formats in dclab

        Returns
        -------
        fmt_dict: dict
            All file formats with names as keys.
            Each item contains the keys "name" (format name),
            "suffix" (saved file suffix), "requires" (Python
            dependencies).

        See Also
        --------
        supported_formats: class-specific file formats
        """
        formats = {}
        for cls in BaseModel.__subclasses__():
            # Register the tensorflow file format
            for fmt in cls.supported_formats():
                formats[fmt["name"]] = {
                    "requirements": fmt["requirements"],
                    "suffix": fmt["suffix"],
                    "class": cls}
        return formats

    @staticmethod
    @abc.abstractmethod
    def supported_formats():
        """List of dictionaries containing model formats

        Returns
        -------
        fmts: list
            Each item contains the keys "name" (format name),
            "suffix" (saved file suffix), "requires" (Python
            dependencies).

        Notes
        -----
        The return value is automatically added to the return value
        of :func:`BaseModel.all_formats`.
        """

    @staticmethod
    @abc.abstractmethod
    def load_bare_model(path):
        """Load an implementation-specific model from a file

        This will set the `self.model` attribute. Make sure that
        the other attributes are set properly as well.
        """

    @staticmethod
    @abc.abstractmethod
    def save_bare_model(path, bare_model, save_format=None):
        """Save an implementation-specific model to a file

        Parameters
        ----------
        path: str or path-like
            Path to store model to
        bare_model: object
            The implementation-specific bare model
        save_format: str
            Must be in `supported_formats`
        """

    @abc.abstractmethod
    def predict(self, ds):
        """Return the probabilities of `self.outputs` for `ds`

        Parameters
        ----------
        ds: dclab.rtdc_dataset.RTDCBase
            Dataset to apply the model to

        Returns
        -------
        ofdict: dict
            Output feature dictionary with features as keys
            and 1d ndarrays as values.

        Notes
        -----
        This function calls :func:`BaseModel.get_dataset_features`
        to obtain the input feature matrix.
        """

    def get_dataset_features(self, ds, dtype=np.float32):
        """Return the dataset features used for inference

        Parameters
        ----------
        ds: dclab.rtdc_dataset.RTDCBase
            Dataset from which to retrieve the feature data
        dtype: dtype
            All features are cast to this dtype

        Returns
        -------
        fdata: 2d ndarray
            2D array of shape (len(ds), len(self.inputs));
            i.e. to access the array containing the first feature,
            for all events, you would do `fdata[:, 0]`.
        """
        fdata = np.zeros((len(ds), len(self.inputs)), dtype=dtype)
        for ii, feat in enumerate(self.inputs):
            fdata[:, ii] = ds[feat]
        return fdata




import numpy as np
import warnings

from dclab.features import emodulus


def test_cell_carrier():
    """Test using table from Christophs script"""
    ch_sizes = [15, 15, 15, 20, 20, 20, 20, 20, 30, 30, 30, 40, 40, 40]
    fl_rates = [0.016, 0.032, 0.048, 0.02, 0.04, 0.06,
                0.08, 0.12, 0.16, 0.24, 0.32, 0.32, 0.40, 0.60]
    temps = [24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24]
    eta_a = [5.8, 4.6, 4.1, 7.1, 5.7, 5.0,
             4.5, 4.0, 5.4, 4.7, 4.3, 5.7, 5.3, 4.6]
    eta_b = [7.5, 5.8, 5.0, 9.4, 7.3, 6.3,
             5.7, 4.9, 6.9, 5.9, 5.3, 7.3, 6.7, 5.8]

    for c, f, t, a in zip(ch_sizes, fl_rates, temps, eta_a):
        eta = emodulus.viscosity.get_viscosity(medium="CellCarrier",
                                               channel_width=c,
                                               flow_rate=f,
                                               temperature=t)
        assert np.allclose(np.round(eta, 1), a)

    for c, f, t, b in zip(ch_sizes, fl_rates, temps, eta_b):
        eta = emodulus.viscosity.get_viscosity(medium="CellCarrier B",
                                               channel_width=c,
                                               flow_rate=f,
                                               temperature=t)
        assert np.allclose(np.round(eta, 1), b)


def test_cellcarrier_range():
    # test values
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(
            medium="CellCarrier B", temperature=15)
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(
            medium="CellCarrier B", temperature=28)
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(medium="CellCarrier B",
                                         temperature=np.arange(-2, 10))
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)


def test_cellcarrierb_range():
    # test values
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(medium="CellCarrier", temperature=15)
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(medium="CellCarrier", temperature=28)
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(medium="CellCarrier",
                                         temperature=np.arange(-2, 10))
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)


def test_water():
    """Test with data from Kestin et al, J. Phys. Chem. 7(3) 1978"""
    ref = np.array([[0, 1791.5],
                    [5, 1519.3],
                    [10, 1307.0],
                    [15, 1138.3],
                    [20, 1002.0],
                    [25, 890.2],
                    [30, 797.3],
                    [35, 719.1],
                    [40, 652.7],
                    ])
    ref[:, 1] *= 1e-3  # uPas to mPas

    res = emodulus.viscosity.get_viscosity(medium="water",
                                           temperature=ref[:, 0])
    assert np.allclose(res, ref[:, 1], rtol=8e-5, atol=0)


def test_water_range():
    # test values
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(medium="water", temperature=-1)
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(medium="water", temperature=41)
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        emodulus.viscosity.get_viscosity(medium="water",
                                         temperature=np.arange(-2, 10))
        assert issubclass(w[-1].category,
                          emodulus.viscosity.TemperatureOutOfRangeWarning)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



def compute_area_exponent(rtdc_ds):
    """Compute area^exp depending on the given user-defined metadata"""
    area_exp = rtdc_ds["area_um"] ** rtdc_ds.config["user"]["exp"]
    return {"area_exp": area_exp}


info = {
    "method": compute_area_exponent,
    "description": "Compute area to the power of exp",
    "feature names": ["area_exp"],
    "features required": ["area_um"],
    "config required": [["user", ["exp"]]],
    "version": "0.1.0",
}



"""Computation of inertia ratio from contour data"""

import numpy as np
import scipy.spatial as ssp


def cont_moments_cv(cont,
                    flt_epsilon=1.19209e-07,
                    dbl_epsilon=2.2204460492503131e-16):
    """Compute the moments of a contour

    The moments are computed in the same way as they are computed
    in OpenCV's `contourMoments` in `moments.cpp`.

    Parameters
    ----------
    cont: array of shape (N,2)
        The contour for which to compute the moments.
    flt_epsilon: float
        The value of ``FLT_EPSILON`` in OpenCV/gcc.
    dbl_epsilon: float
        The value of ``DBL_EPSILON`` in OpenCV/gcc.

    Returns
    -------
    moments: dict
        A dictionary of moments. If the moment `m00` is smaller
        than half of `flt_epsilon`, `None` is returned.
    """
    # Make sure we have no unsigned integers
    if np.issubdtype(cont.dtype, np.unsignedinteger):
        cont = cont.astype(int)

    xi = cont[:, 0]
    yi = cont[:, 1]

    xi_1 = np.roll(xi, -1)
    yi_1 = np.roll(yi, -1)

    xi_12 = xi_1**2
    yi_12 = yi_1**2

    xi2 = xi**2
    yi2 = yi**2

    dxy = xi_1 * yi - xi * yi_1

    xii_1 = xi_1 + xi
    yii_1 = yi_1 + yi

    a00 = np.sum(dxy)
    a10 = np.sum(dxy * xii_1)
    a01 = np.sum(dxy * yii_1)
    a20 = np.sum(dxy * (xi_1 * xii_1 + xi2))
    a11 = np.sum(dxy * (xi_1 * (yii_1 + yi_1) + xi * (yii_1 + yi)))
    a02 = np.sum(dxy * (yi_1 * yii_1 + yi2))
    a30 = np.sum(dxy * xii_1 * (xi_12 + xi2))
    a03 = np.sum(dxy * yii_1 * (yi_12 + yi2))
    a21 = np.sum(dxy * (xi_12 * (3 * yi_1 + yi) + 2 *
                        xi * xi_1 * yii_1 + xi2 * (yi_1 + 3 * yi)))
    a12 = np.sum(dxy * (yi_12 * (3 * xi_1 + xi) + 2 *
                        yi * yi_1 * xii_1 + yi2 * (xi_1 + 3 * xi)))

    if abs(a00) > flt_epsilon:
        db1_2 = 0.5
        db1_6 = 0.16666666666666666666666666666667
        db1_12 = 0.083333333333333333333333333333333
        db1_24 = 0.041666666666666666666666666666667
        db1_20 = 0.05
        db1_60 = 0.016666666666666666666666666666667

        if a00 < 0:
            db1_2 *= -1
            db1_6 *= -1
            db1_12 *= -1
            db1_24 *= -1
            db1_20 *= -1
            db1_60 *= -1

        m = dict(m00=a00 * db1_2,
                 m10=a10 * db1_6,
                 m01=a01 * db1_6,
                 m20=a20 * db1_12,
                 m11=a11 * db1_24,
                 m02=a02 * db1_12,
                 m30=a30 * db1_20,
                 m21=a21 * db1_60,
                 m12=a12 * db1_60,
                 m03=a03 * db1_20,
                 )

        if m["m00"] > dbl_epsilon:
            # Center of gravity
            cx = m["m10"]/m["m00"]
            cy = m["m01"]/m["m00"]
        else:
            cx = 0
            cy = 0

        # central second order moments
        m["mu20"] = m["m20"] - m["m10"]*cx
        m["mu11"] = m["m11"] - m["m10"]*cy
        m["mu02"] = m["m02"] - m["m01"]*cy

        m["mu30"] = m["m30"] - cx*(3*m["mu20"] + cx*m["m10"])
        m["mu21"] = m["m21"] - cx*(2*m["mu11"] + cx*m["m01"]) - cy*m["mu20"]
        m["mu12"] = m["m12"] - cy*(2*m["mu11"] + cy*m["m10"]) - cx*m["mu02"]
        m["mu03"] = m["m03"] - cy*(3*m["mu02"] + cy*m["m01"])
        return m
    else:
        return None


def get_inert_ratio_cvx(cont):
    """Compute the inertia ratio of the convex hull of a contour

    The inertia ratio is computed from the central second order of moments
    along x (mu20) and y (mu02) via `sqrt(mu20/mu02)`.

    Parameters
    ----------
    cont: ndarray or list of ndarrays of shape (N,2)
        A 2D array that holds the contour of an event (in pixels)
        e.g. obtained using `mm.contour` where  `mm` is an instance
        of `RTDCBase`. The first and second columns of `cont`
        correspond to the x- and y-coordinates of the contour.

    Returns
    -------
    inert_ratio_cvx: float or ndarray of size N
        The inertia ratio of the contour's convex hull

    Notes
    -----
    The contour moments mu20 and mu02 are computed the same way they
    are computed in OpenCV's `moments.cpp`.


    See Also
    --------
    get_inert_ratio_raw: Compute inertia ratio of a raw contour


    References
    ----------
    - `<https://en.wikipedia.org/wiki/Image_moment#Central_moments>`__
    - `<https://github.com/opencv/opencv/blob/
      f81370232a651bdac5042efe907bcaa50a66c487/modules/imgproc/src/
      moments.cpp#L93>`__
    """
    if isinstance(cont, np.ndarray):
        # If cont is an array, it is not a list of contours,
        # because contours can have different lengths.
        cont = [cont]
        ret_list = False
    else:
        ret_list = True

    length = len(cont)

    inert_ratio_cvx = np.zeros(length, dtype=float) * np.nan

    for ii in range(length):
        try:
            chull = ssp.ConvexHull(cont[ii])
        except ssp.qhull.QhullError:
            pass
        else:
            hull = cont[ii][chull.vertices, :]
            inert_ratio_cvx[ii] = get_inert_ratio_raw(hull)

    if not ret_list:
        inert_ratio_cvx = inert_ratio_cvx[0]

    return inert_ratio_cvx


def get_inert_ratio_prnc(cont):
    """Compute principal inertia ratio of a contour

    The principal inertia ratio is rotation-invariant, which
    makes it applicable to reservoir measurements where e.g.
    cells are not aligned with the channel.

    Parameters
    ----------
    cont: ndarray or list of ndarrays of shape (N,2)
        A 2D array that holds the contour of an event (in pixels)
        e.g. obtained using `mm.contour` where  `mm` is an instance
        of `RTDCBase`. The first and second columns of `cont`
        correspond to the x- and y-coordinates of the contour.

    Returns
    -------
    inert_ratio_prnc: float or ndarray of size N
        The principal inertia ratio of the contour
    """
    if isinstance(cont, np.ndarray):
        # If cont is an array, it is not a list of contours,
        # because contours can have different lengths.
        cont = [cont]
        ret_list = False
    else:
        ret_list = True

    length = len(cont)

    inert_ratio_prnc = np.zeros(length, dtype=float) * np.nan

    for ii in range(length):
        moments = cont_moments_cv(cont[ii])
        if moments is not None:
            # orientation of the contour
            orient = 0.5 * np.arctan2(2 * moments['mu11'],
                                      moments['mu02'] - moments['mu20'])
            # require floating point array (only copy if necessary)
            cc = np.array(cont[ii], dtype=float, copy=False)
            # rotate contour
            rho = np.sqrt(cc[:, 0]**2 + cc[:, 1]**2)
            phi = np.arctan2(cc[:, 1], cc[:, 0]) + orient + np.pi / 2
            cc[:, 0] = rho * np.cos(phi)
            cc[:, 1] = rho * np.sin(phi)
            # compute inertia ratio of rotated contour
            mprnc = cont_moments_cv(cc)
            inert_ratio_prnc[ii] = np.sqrt(mprnc["mu20"] / mprnc["mu02"])

    if not ret_list:
        inert_ratio_prnc = inert_ratio_prnc[0]

    return inert_ratio_prnc


def get_inert_ratio_raw(cont):
    """Compute the inertia ratio of a contour

    The inertia ratio is computed from the central second order of moments
    along x (mu20) and y (mu02) via `sqrt(mu20/mu02)`.

    Parameters
    ----------
    cont: ndarray or list of ndarrays of shape (N,2)
        A 2D array that holds the contour of an event (in pixels)
        e.g. obtained using `mm.contour` where  `mm` is an instance
        of `RTDCBase`. The first and second columns of `cont`
        correspond to the x- and y-coordinates of the contour.

    Returns
    -------
    inert_ratio_raw: float or ndarray of size N
        The inertia ratio of the contour

    Notes
    -----
    The contour moments mu20 and mu02 are computed the same way they
    are computed in OpenCV's `moments.cpp`.

    See Also
    --------
    get_inert_ratio_cvx: Compute inertia ratio of the convex hull of
                         a contour

    References
    ----------
    - `<https://en.wikipedia.org/wiki/Image_moment#Central_moments>`__
    - `<https://github.com/opencv/opencv/blob/
      f81370232a651bdac5042efe907bcaa50a66c487/modules/imgproc/src/
      moments.cpp#L93>`__
    """
    if isinstance(cont, np.ndarray):
        # If cont is an array, it is not a list of contours,
        # because contours can have different lengths.
        cont = [cont]
        ret_list = False
    else:
        ret_list = True

    length = len(cont)

    inert_ratio_raw = np.zeros(length, dtype=float) * np.nan

    for ii in range(length):
        moments = cont_moments_cv(cont[ii])
        if moments is not None:
            inert_ratio_raw[ii] = np.sqrt(moments["mu20"]/moments["mu02"])

    if not ret_list:
        inert_ratio_raw = inert_ratio_raw[0]

    return inert_ratio_raw


def get_tilt(cont):
    """Compute tilt of raw contour relative to channel axis

    Parameters
    ----------
    cont: ndarray or list of ndarrays of shape (N,2)
        A 2D array that holds the contour of an event (in pixels)
        e.g. obtained using `mm.contour` where  `mm` is an instance
        of `RTDCBase`. The first and second columns of `cont`
        correspond to the x- and y-coordinates of the contour.

    Returns
    -------
    tilt: float or ndarray of size N
        Tilt of the contour in the interval [0, PI/2]

    References
    ----------
    - `<https://en.wikipedia.org/wiki/Image_moment#Examples_2>`__
    """
    if isinstance(cont, np.ndarray):
        # If cont is an array, it is not a list of contours,
        # because contours can have different lengths.
        cont = [cont]
        ret_list = False
    else:
        ret_list = True

    length = len(cont)

    tilt = np.zeros(length, dtype=float) * np.nan

    for ii in range(length):
        moments = cont_moments_cv(cont[ii])
        if moments is not None:
            # orientation of the contour
            oii = 0.5 * np.arctan2(2 * moments['mu11'],
                                   moments['mu02'] - moments['mu20'])
            # +PI/2 because relative to channel axis
            tilt[ii] = oii + np.pi/2

    # restrict to interval [0,PI/2]
    tilt = np.mod(tilt, np.pi)
    # avoid nans
    valid = ~np.isnan(tilt)
    wrap = tilt[valid]
    wrap[wrap > np.pi/2] -= np.pi
    tilt[valid] = wrap
    tilt = np.abs(tilt)

    if not ret_list:
        tilt = tilt[0]

    return tilt



"""A wrapper around R with the lme4 package"""
# flake8: noqa: F401
from . import rlibs, rsetup, wrapr
from .wrapr import Rlme4, bootstrapped_median_distributions
from .rsetup import get_r_path, get_r_version, install_lme4, set_r_path



"""Test filter hierarchies"""
import numpy as np
import pytest

from dclab import new_dataset
from dclab.rtdc_dataset import fmt_hierarchy, RTDCWriter

from helper_methods import example_data_dict, retrieve_data


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_config_calculation():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                "emodulus medium": "CellCarrier",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5
                                }
    ch = new_dataset(ds)
    assert np.allclose(ch["emodulus"], ds["emodulus"], equal_nan=True)
    ds.config["calculation"]["emodulus temperature"] = 24.0
    assert np.allclose(ch["emodulus"], ds["emodulus"], equal_nan=True)
    # the user still has to call `apply_filter` to update the config:
    assert not ch.config["calculation"]["emodulus temperature"] == 24.0
    ch.apply_filter()
    assert ch.config["calculation"]["emodulus temperature"] == 24.0


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hierarchy_shape_contour():
    ds = new_dataset(retrieve_data("fmt-hdf5_image-bg_2020.zip"))
    assert ds["contour"].shape == (5, np.nan, 2)
    ds.filter.manual[0] = False
    ch = new_dataset(ds)
    assert ch["contour"].shape == (4, np.nan, 2)
    assert len(ch["contour"]) == 4


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hierarchy_shape_image():
    ds = new_dataset(retrieve_data("fmt-hdf5_image-bg_2020.zip"))
    ds.filter.manual[0] = False
    ch = new_dataset(ds)
    assert "image" in ch.features_innate
    assert len(ch["image"]) == 4
    assert ch["image"].shape == (4, 80, 250)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hierarchy_shape_mask():
    ds = new_dataset(retrieve_data("fmt-hdf5_image-bg_2020.zip"))
    ds.filter.manual[0] = False
    ch = new_dataset(ds)
    assert "mask" in ch.features_innate
    assert len(ch["mask"]) == 4
    assert ch["mask"].shape == (4, 80, 250)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hierarchy_shape_trace():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    assert len(ds) == 7
    ds.filter.manual[0] = False
    ds.filter.manual[1] = False
    ch = new_dataset(ds)
    assert len(ch) == 5
    assert "trace" in ch.features_innate
    assert ch["trace"].shape == (6, 5, 177)
    assert ch["trace"]["fl1_raw"].shape == (5, 177)
    assert ch["trace"]["fl1_raw"][0].shape == (177,)
    assert len(ch["trace"]) == 6
    assert len(ch["trace"]["fl1_raw"]) == 5
    assert len(ch["trace"]["fl1_raw"][0]) == 177


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_event_count():
    pytest.importorskip("nptdms")
    tdms_path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    ds = new_dataset(tdms_path)
    ds.filter.manual[0] = False
    ch = new_dataset(ds)
    assert ds.config["experiment"]["event count"] == len(ds)
    assert ch.config["experiment"]["event count"] == len(ch)
    assert len(ds) == len(ch) + 1


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_feat_contour():
    path = retrieve_data("fmt-hdf5_fl_2017.zip")
    ds = new_dataset(path)
    ds.filter.manual[0] = False
    ds.filter.manual[2] = False
    ch = new_dataset(ds)
    assert np.all(ch["contour"][0] == ds["contour"][1])
    assert np.all(ch["contour"][1] == ds["contour"][3])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_feat_image():
    path = retrieve_data("fmt-hdf5_fl_2017.zip")
    ds = new_dataset(path)
    ds.filter.manual[0] = False
    ds.filter.manual[2] = False
    ch = new_dataset(ds)
    assert np.all(ch["image"][0] == ds["image"][1])
    assert np.all(ch["image"][1] == ds["image"][3])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_feat_image_bg():
    path = retrieve_data("fmt-hdf5_fl_2017.zip")
    # add a fake image_bg column
    with RTDCWriter(path) as hw:
        image_bg = hw.h5file["events"]["image"][:] // 2
        hw.store_feature("image_bg", image_bg)
    ds = new_dataset(path)
    ds.filter.manual[0] = False
    ds.filter.manual[2] = False
    ch = new_dataset(ds)
    assert np.all(ch["image_bg"][0] == ds["image_bg"][1])
    assert np.all(ch["image_bg"][1] == ds["image_bg"][3])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_feat_mask():
    path = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    ds = new_dataset(path)
    ds.filter.manual[0] = False
    ds.filter.manual[2] = False
    ch = new_dataset(ds)
    assert np.all(ch["mask"][0] == ds["mask"][1])
    assert np.all(ch["mask"][1] == ds["mask"][3])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_feat_trace():
    path = retrieve_data("fmt-hdf5_fl_2017.zip")
    ds = new_dataset(path)
    ds.filter.manual[0] = False
    ds.filter.manual[2] = False
    ch = new_dataset(ds)
    assert np.all(ch["trace"]["fl1_median"][0]
                  == ds["trace"]["fl1_median"][1])
    assert np.all(ch["trace"]["fl1_median"][1]
                  == ds["trace"]["fl1_median"][3])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_features():
    path = retrieve_data("fmt-hdf5_fl_2017.zip")
    ds = new_dataset(path)
    ch = new_dataset(ds)
    assert ds.features == ch.features
    assert ds.features_innate == ch.features_innate
    assert ds.features_loaded == ch.features_loaded
    assert ds.features_scalar == ch.features_scalar


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_feature_contained():
    pytest.importorskip("nptdms")
    path_im = retrieve_data("fmt-tdms_fl-image_2016.zip")
    ds_im = new_dataset(path_im)
    path_no_im = retrieve_data("fmt-tdms_2fl-no-image_2017.zip")
    ds_no_im = new_dataset(path_no_im)

    assert "image" in ds_im
    assert "mask" in ds_im
    assert "image" not in ds_no_im
    assert "mask" not in ds_no_im


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_features_loaded():
    path = retrieve_data("fmt-hdf5_fl_2017.zip")
    ds = new_dataset(path)
    ch = new_dataset(ds)
    assert "volume" in ch.features
    assert "volume" not in ch.features_loaded
    assert "volume" not in ch.features_innate
    # compute volume, now it should be loaded
    ch["volume"]
    assert "volume" in ch.features_loaded
    assert "volume" in ds.features_loaded


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hierarchy_from_tdms():
    pytest.importorskip("nptdms")
    tdms_path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    ds1 = new_dataset(tdms_path)
    ds2 = new_dataset(ds1)

    ds1.filter.manual[0] = False
    ds2.apply_filter()
    assert ds2.filter.all.shape[0] == ds1.filter.all.shape[0] - 1
    assert ds2["area_um"][0] == ds1["area_um"][1]


def test_index_deep_contour():
    data = example_data_dict(42, keys=["area_um", "contour", "deform"])
    ds = new_dataset(data)
    ds.filter.manual[3] = False
    c1 = new_dataset(ds)
    c1.filter.manual[1] = False
    c2 = new_dataset(c1)
    assert np.all(c2["contour"][3] == ds["contour"][5])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("feat", ["image", "image_bg", "mask"])
@pytest.mark.parametrize("idxs", [slice(0, 3), np.arange(3),
                                  [0, 1, 2], [True, True, True, False]])
def test_index_slicing(feat, idxs):
    data = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    ds = new_dataset(data)
    ds.filter.manual[2] = False
    ch = new_dataset(ds)

    ds_feat = ds[feat][np.array([0, 1, 3])]
    ch_feat = ch[feat]

    assert np.all(ch_feat[idxs] == ds_feat)


@pytest.mark.parametrize("feat", ["image", "mask"])
@pytest.mark.parametrize("idxs", [slice(0, 3), np.arange(3),
                                  [0, 1, 2], [False]+42*[True]])
def test_index_slicing_tdms_fails(feat, idxs):
    """The tdms-file format does not support slice/array indexing"""
    pytest.importorskip("nptdms")
    data = retrieve_data("fmt-tdms_fl-image_2016.zip")
    ds = new_dataset(data)
    ds.filter.manual[2] = False
    ch = new_dataset(ds)

    with pytest.raises(NotImplementedError, match="scalar integers"):
        ch[feat][idxs]


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
@pytest.mark.parametrize("idxs", [slice(0, 3), np.arange(3),
                                  [0, 1, 2], [True, True, True, False]])
def test_index_slicing_trace(idxs):
    data = retrieve_data("fmt-hdf5_fl_2017.zip")
    ds = new_dataset(data)
    ds.filter.manual[2] = False
    ch = new_dataset(ds)

    ds_feat = ds["trace"]["fl1_median"][np.array([0, 1, 3])]
    ch_feat = ch["trace"]["fl1_median"]

    assert np.all(ch_feat[idxs] == ds_feat)


def test_manual_exclude():
    data = example_data_dict(42, keys=["area_um", "deform"])
    p = new_dataset(data)
    c1 = new_dataset(p)
    c2 = new_dataset(c1)
    c3 = new_dataset(c2)
    c1.filter.manual[0] = False
    c2.apply_filter()
    c2.filter.manual[1] = False
    c3.apply_filter()

    # simple exclusion of few events
    assert len(c3) == len(p) - 2

    # removing same event in parent removes the event from the
    # child altogether, including the manual filter
    c3.filter.manual[0] = False
    c2.filter.manual[0] = False
    c3.apply_filter()
    assert np.alltrue(c3.filter.manual)

    # reinserting the event in the parent, retrieves back
    # the manual filter in the child
    c2.filter.manual[0] = True
    c3.apply_filter()
    assert not c3.filter.manual[0]


def test_manual_exclude_parent_changed():
    data = example_data_dict(42, keys=["area_um", "tilt"])
    p = new_dataset(data)
    p.filter.manual[4] = False
    c = new_dataset(p)
    c.filter.manual[5] = False
    c.apply_filter()
    p.config["filtering"]["tilt min"] = 0
    p.config["filtering"]["tilt max"] = .5
    p.apply_filter()
    assert np.sum(p.filter.all) == 21
    # size of child is directly determined from parent
    assert len(c) == 21
    # filters have not yet been updated
    assert len(c.filter.all) == 41
    assert c.filter.parent_changed
    # the initially excluded event
    assert c.filter.retrieve_manual_indices(c) == [6]

    # try to change the excluded events
    try:
        c.filter.apply_manual_indices(c, [1, 2])
    except fmt_hierarchy.HierarchyFilterError:
        pass
    else:
        assert False, "expected HierarchyFilterError"

    # this can be resolved by applying the filter
    c.apply_filter()
    c.filter.apply_manual_indices(c, [1, 2])
    assert c.filter.retrieve_manual_indices(c) == [1, 2]


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_same_hash_different_identifier():
    pytest.importorskip("nptdms")
    tdms_path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    ds1 = new_dataset(tdms_path)
    ds1.filter.manual[0] = False
    ch1 = new_dataset(ds1)
    ch2 = new_dataset(ds1)
    assert len(ch1) == len(ds1) - 1
    assert ch1.hash == ch2.hash
    assert ch1.identifier != ch2.identifier


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""RT-DC dataset core classes and methods"""
import abc
import random
import warnings

import numpy as np

from .. import definitions as dfn
from .. import downsampling
from ..polygon_filter import PolygonFilter
from .. import kde_methods

from .feat_anc_core import AncillaryFeature, FEATURES_RAPID
from .export import Export
from .filter import Filter


class LogTransformWarning(UserWarning):
    pass


class RTDCBase(abc.ABC):
    def __init__(self, identifier=None):
        """RT-DC measurement base class

        Notes
        -----
        Besides the filter arrays for each data feature, there is a manual
        boolean filter array ``RTDCBase.filter.manual`` that can be edited
        by the user - a boolean value of ``False`` means that the event is
        excluded from all computations.
        """
        #: Dataset format (derived from class name)
        self.format = self.__class__.__name__.split("_")[-1].lower()

        self._polygon_filter_ids = []
        # Events have the feature name as keys and contain nD ndarrays.
        self._events = {}
        # Ancillaries have the feature name as keys and a
        # tuple containing feature and hash as value.
        self._ancillaries = {}
        # Temporary features are defined by the user ad hoc at runtime.
        self._usertemp = {}
        #: Configuration of the measurement
        self.config = None
        #: Export functionalities; instance of
        #: :class:`dclab.rtdc_dataset.export.Export`.
        self.export = Export(self)
        # The filtering class is initialized with self._init_filters
        #: Filtering functionalities; instance of
        #: :class:`dclab.rtdc_dataset.filter.Filter`.
        self.filter = None
        #: Dictionary of log files. Each log file is a list of strings
        #: (one string per line).
        self.logs = {}
        #: Title of the measurement
        self.title = None
        #: Path or DCOR identifier of the dataset (set to "none"
        #: for :class:`RTDC_Dict`)
        self.path = None
        # Unique identifier
        if identifier is None:
            # Generate a unique identifier for this dataset
            rhex = [random.choice('0123456789abcdef') for _n in range(7)]
            self._identifier = "mm-{}_{}".format(self.format, "".join(rhex))
        else:
            self._identifier = identifier

    def __contains__(self, key):
        ct = False
        if key in self._events or key in self._usertemp:
            ct = True
        else:
            # Check ancillary features data
            if key in self._ancillaries:
                # already computed
                ct = True
            elif key in AncillaryFeature.feature_names:
                # get all instance of AncillaryFeature that
                # compute the feature `key`
                instlist = AncillaryFeature.get_instances(key)
                for inst in instlist:
                    if inst.is_available(self):
                        # to be computed
                        ct = True
                        break
        return ct

    def __getitem__(self, key):
        if key in self._events:
            return self._events[key]
        elif key in self._usertemp:
            return self._usertemp[key]
        # Try to find the feature in the ancillary features
        # (see feat_anc_core submodule for more information).
        # These features are cached in `self._ancillaries`.
        ancol = AncillaryFeature.available_features(self)
        if key in ancol:
            # The feature is available.
            anhash = ancol[key].hash(self)
            if (key in self._ancillaries and
                    self._ancillaries[key][0] == anhash):
                # Use cached value
                data = self._ancillaries[key][1]
            else:
                # Compute new value
                data_dict = ancol[key].compute(self)
                for okey in data_dict:
                    # Store computed value in `self._ancillaries`.
                    self._ancillaries[okey] = (anhash, data_dict[okey])
                data = data_dict[key]
            return data
        else:
            raise KeyError(f"Feature '{key}' does not exist in {self}!")

    def __iter__(self):
        """An iterator over all valid scalar features"""
        mycols = []
        for col in self._feature_candidates:
            if col in self:
                mycols.append(col)
        mycols.sort()
        for col in mycols:
            yield col

    def __len__(self):
        keys = list(self._events.keys())
        keys.sort()
        for kk in keys:
            length = len(self._events[kk])
            if length:
                return length
        else:
            msg = "Could not determine size of dataset '{}'.".format(self)
            raise ValueError(msg)

    def __repr__(self):
        repre = "<{} '{}' at {}".format(self.__class__.__name__,
                                        self.identifier,
                                        hex(id(self)))
        if self.path != "none":
            repre += " ({})>".format(self.path)
        return repre

    @staticmethod
    def _apply_scale(a, scale, feat):
        """Helper function for transforming an aray to log-scale

        Parameters
        ----------
        a: np.ndarray
            Input array
        scale: str
            If set to "log", take the logarithm of `a`; if set to
            "linear" return `a` unchanged.
        feat: str
            Feature name (required for debugging)

        Returns
        -------
        b: np.ndarray
            The scaled array

        Notes
        -----
        If the scale is not "linear", then a new array is returned.
        All warnings are suppressed when computing `np.log(a)`, as
        `a` may have negative or nan values.
        """
        if scale == "linear":
            b = a
        elif scale == "log":
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                b = np.log(a)
                if len(w):
                    # Tell the user that the log-transformation issued
                    # a warning.
                    warnings.warn("Invalid values encounterd in np.log "
                                  "while scaling feature '{}'!".format(feat))
        else:
            raise ValueError("`scale` must be either 'linear' or 'log', "
                             + "got '{}'!".format(scale))
        return b

    @staticmethod
    def get_kde_spacing(a, scale="linear", method=kde_methods.bin_width_doane,
                        method_kw=None, feat="undefined", ret_scaled=False):
        """Convenience function for computing the contour spacing

        Parameters
        ----------
        a: ndarray
            feature data
        scale: str
            how the data should be scaled ("log" or "linear")
        method: callable
            KDE method to use (see `kde_methods` submodule)
        method_kw: dict
            keyword arguments to `method`
        feat: str
            feature name for debugging
        ret_scaled: bool
            whether or not to return the scaled array of `a`
        """
        if method_kw is None:
            method_kw = {}
        # Apply scale (no change for linear scale)
        asc = RTDCBase._apply_scale(a, scale, feat)
        # Apply multiplicator
        acc = method(asc, **method_kw)
        if ret_scaled:
            return acc, asc
        else:
            return acc

    @property
    def _feature_candidates(self):
        """List of feature candidates for this dataset

        Use with caution! Features in this list might not actually
        be available. Always check against `__contains__`.
        """
        feats = list(self._events.keys())
        feats += list(self._usertemp.keys())
        feats += list(AncillaryFeature.feature_names)
        feats = sorted(set(feats))
        # exclude non-standard features
        featsv = [ff for ff in feats if dfn.feature_exists(ff)]
        return featsv

    @property
    def _filter(self):
        """return the current filter boolean array"""
        warnings.warn("RTDCBase._filter is deprecated. Please use "
                      + "RTDCBase.filter.all instead.",
                      DeprecationWarning)
        return self.filter.all

    @property
    def _plot_filter(self):
        raise NotImplementedError(
            "RTDCBase._plot_filter has been removed in dclab 0.16.0. "
            + "Please use the output of RTDCBase.downsample_scatter "
            + "with the argument ret_mask instead.")

    def _init_filters(self):
        #: Filtering functionalities (this is an instance of
        #: :class:`dclab.rtdc_dataset.filter.Filter`.
        self.filter = Filter(self)
        self.reset_filter()

    @property
    def identifier(self):
        """Unique (unreproducible) identifier"""
        return self._identifier

    @property
    def features(self):
        """All available features"""
        mycols = []
        for col in self._feature_candidates:
            if col in self:
                mycols.append(col)
        mycols.sort()
        return mycols

    @property
    def features_innate(self):
        """All features excluding ancillary or temporary features"""
        innate = [ft for ft in self.features if ft in self._events]
        return innate

    @property
    def features_loaded(self):
        """All features that have been computed

        This includes ancillary features and temporary features.

        Notes
        -----
        Features that are computationally cheap to compute are
        always included. They are defined in
        :const:`dclab.rtdc_dataset.feat_anc_core.FEATURES_RAPID`.
        """
        features_innate = self.features_innate
        features_loaded = []
        for feat in self.features:
            if (feat in features_innate
                or feat in FEATURES_RAPID
                or feat in self._usertemp
                    or feat in self._ancillaries):
                # Note that there is no hash checking here for
                # ancillary features. This might be interesting
                # only in rare cases.
                features_loaded.append(feat)
        return features_loaded

    @property
    def features_scalar(self):
        """All scalar features available"""
        sclr = [ft for ft in self.features if dfn.scalar_feature_exists(ft)]
        return sclr

    @property
    @abc.abstractmethod
    def hash(self):
        """Reproducible dataset hash (defined by derived classes)"""

    def apply_filter(self, force=None):
        """Compute the filters for the dataset"""
        if force is None:
            force = []
        self.filter.update(rtdc_ds=self, force=force)

    def get_downsampled_scatter(self, xax="area_um", yax="deform",
                                downsample=0, xscale="linear",
                                yscale="linear", remove_invalid=False,
                                ret_mask=False):
        """Downsampling by removing points at dense locations

        Parameters
        ----------
        xax: str
            Identifier for x axis (e.g. "area_um", "aspect", "deform")
        yax: str
            Identifier for y axis
        downsample: int
            Number of points to draw in the down-sampled plot.
            This number is either

            - >=1: exactly downsample to this number by randomly adding
                   or removing points
            - 0  : do not perform downsampling
        xscale: str
            If set to "log", take the logarithm of the x-values before
            performing downsampling. This is useful when data are are
            displayed on a log-scale. Defaults to "linear".
        yscale: str
            See `xscale`.
        remove_invalid: bool
            Remove nan and inf values before downsampling; if set to
            `True`, the actual number of samples returned might be
            smaller than `downsample` due to infinite or nan values
            (e.g. due to logarithmic scaling).
        ret_mask: bool
            If set to `True`, returns a boolean array of length
            `len(self)` where `True` values identify the filtered
            data.

        Returns
        -------
        xnew, xnew: 1d ndarray of lenght `N`
            Filtered data; `N` is either identical to `downsample`
            or smaller (if `remove_invalid==True`)
        mask: 1d boolean array of length `len(RTDCBase)`
            Array for identifying the downsampled data points
        """
        if downsample < 0:
            raise ValueError("`downsample` must be zero or positive!")

        downsample = int(downsample)
        xax = xax.lower()
        yax = yax.lower()

        # Get data
        x = self[xax][self.filter.all]
        y = self[yax][self.filter.all]

        # Apply scale (no change for linear scale)
        xs = RTDCBase._apply_scale(x, xscale, xax)
        ys = RTDCBase._apply_scale(y, yscale, yax)

        _, _, idx = downsampling.downsample_grid(xs, ys,
                                                 samples=downsample,
                                                 remove_invalid=remove_invalid,
                                                 ret_idx=True)

        if ret_mask:
            # Mask is a boolean array of len(self)
            mask = np.zeros(len(self), dtype=bool)
            mids = np.where(self.filter.all)[0]
            mask[mids] = idx
            return x[idx], y[idx], mask
        else:
            return x[idx], y[idx]

    def get_kde_contour(self, xax="area_um", yax="deform", xacc=None,
                        yacc=None, kde_type="histogram", kde_kwargs=None,
                        xscale="linear", yscale="linear"):
        """Evaluate the kernel density estimate for contour plots

        Parameters
        ----------
        xax: str
            Identifier for X axis (e.g. "area_um", "aspect", "deform")
        yax: str
            Identifier for Y axis
        xacc: float
            Contour accuracy in x direction
        yacc: float
            Contour accuracy in y direction
        kde_type: str
            The KDE method to use
        kde_kwargs: dict
            Additional keyword arguments to the KDE method
        xscale: str
            If set to "log", take the logarithm of the x-values before
            computing the KDE. This is useful when data are
            displayed on a log-scale. Defaults to "linear".
        yscale: str
            See `xscale`.

        Returns
        -------
        X, Y, Z : coordinates
            The kernel density Z evaluated on a rectangular grid (X,Y).
        """
        if kde_kwargs is None:
            kde_kwargs = {}
        xax = xax.lower()
        yax = yax.lower()
        kde_type = kde_type.lower()
        if kde_type not in kde_methods.methods:
            raise ValueError("Not a valid kde type: {}!".format(kde_type))

        # Get data
        x = self[xax][self.filter.all]
        y = self[yax][self.filter.all]

        xacc_sc, xs = RTDCBase.get_kde_spacing(
            a=x,
            feat=xax,
            scale=xscale,
            method=kde_methods.bin_width_doane,
            ret_scaled=True)

        yacc_sc, ys = RTDCBase.get_kde_spacing(
            a=y,
            feat=yax,
            scale=yscale,
            method=kde_methods.bin_width_doane,
            ret_scaled=True)

        if xacc is None:
            xacc = xacc_sc / 5

        if yacc is None:
            yacc = yacc_sc / 5

        # Ignore infs and nans
        bad = kde_methods.get_bad_vals(xs, ys)
        xc = xs[~bad]
        yc = ys[~bad]

        xnum = int(np.ceil((xc.max() - xc.min()) / xacc))
        ynum = int(np.ceil((yc.max() - yc.min()) / yacc))

        xlin = np.linspace(xc.min(), xc.max(), xnum, endpoint=True)
        ylin = np.linspace(yc.min(), yc.max(), ynum, endpoint=True)

        xmesh, ymesh = np.meshgrid(xlin, ylin, indexing="ij")

        kde_fct = kde_methods.methods[kde_type]
        if len(x):
            density = kde_fct(events_x=xs, events_y=ys,
                              xout=xmesh, yout=ymesh,
                              **kde_kwargs)
        else:
            density = np.array([])

        # Convert mesh back to linear scale if applicable
        if xscale == "log":
            xmesh = np.exp(xmesh)
        if yscale == "log":
            ymesh = np.exp(ymesh)

        return xmesh, ymesh, density

    def get_kde_scatter(self, xax="area_um", yax="deform", positions=None,
                        kde_type="histogram", kde_kwargs=None, xscale="linear",
                        yscale="linear"):
        """Evaluate the kernel density estimate for scatter plots

        Parameters
        ----------
        xax: str
            Identifier for X axis (e.g. "area_um", "aspect", "deform")
        yax: str
            Identifier for Y axis
        positions: list of two 1d ndarrays or ndarray of shape (2, N)
            The positions where the KDE will be computed. Note that
            the KDE estimate is computed from the points that
            are set in `self.filter.all`.
        kde_type: str
            The KDE method to use
        kde_kwargs: dict
            Additional keyword arguments to the KDE method
        xscale: str
            If set to "log", take the logarithm of the x-values before
            computing the KDE. This is useful when data are are
            displayed on a log-scale. Defaults to "linear".
        yscale: str
            See `xscale`.

        Returns
        -------
        density : 1d ndarray
            The kernel density evaluated for the filtered data points.
        """
        if kde_kwargs is None:
            kde_kwargs = {}
        xax = xax.lower()
        yax = yax.lower()
        kde_type = kde_type.lower()
        if kde_type not in kde_methods.methods:
            raise ValueError("Not a valid kde type: {}!".format(kde_type))

        # Get data
        x = self[xax][self.filter.all]
        y = self[yax][self.filter.all]

        # Apply scale (no change for linear scale)
        xs = RTDCBase._apply_scale(x, xscale, xax)
        ys = RTDCBase._apply_scale(y, yscale, yax)

        if positions is None:
            posx = None
            posy = None
        else:
            posx = RTDCBase._apply_scale(positions[0], xscale, xax)
            posy = RTDCBase._apply_scale(positions[1], yscale, yax)

        kde_fct = kde_methods.methods[kde_type]
        if len(x):
            density = kde_fct(events_x=xs, events_y=ys,
                              xout=posx, yout=posy,
                              **kde_kwargs)
        else:
            density = np.array([])

        return density

    def polygon_filter_add(self, filt):
        """Associate a Polygon Filter with this instance

        Parameters
        ----------
        filt: int or instance of `PolygonFilter`
            The polygon filter to add
        """
        if not isinstance(filt, (PolygonFilter, int, float)):
            msg = "`filt` must be a number or instance of PolygonFilter!"
            raise ValueError(msg)

        if isinstance(filt, PolygonFilter):
            uid = filt.unique_id
        else:
            uid = int(filt)
        # append item
        self.config["filtering"]["polygon filters"].append(uid)

    def polygon_filter_rm(self, filt):
        """Remove a polygon filter from this instance

        Parameters
        ----------
        filt: int or instance of `PolygonFilter`
            The polygon filter to remove
        """
        if not isinstance(filt, (PolygonFilter, int, float)):
            msg = "`filt` must be a number or instance of PolygonFilter!"
            raise ValueError(msg)

        if isinstance(filt, PolygonFilter):
            uid = filt.unique_id
        else:
            uid = int(filt)
        # remove item
        self.config["filtering"]["polygon filters"].remove(uid)

    def reset_filter(self):
        """Reset the current filter"""
        # reset filter instance
        self.filter.reset()
        # reset configuration
        # remember hierarchy parent
        hp = self.config["filtering"]["hierarchy parent"]
        self.config._init_default_filter_values()
        self.config["filtering"]["hierarchy parent"] = hp



"""Scale conversion applicable to a linear elastic model"""

import warnings

import numpy as np


def convert(area_um, deform, channel_width_in, channel_width_out,
            emodulus=None, flow_rate_in=None, flow_rate_out=None,
            viscosity_in=None, viscosity_out=None, inplace=False):
    """convert area-deformation-emodulus triplet

    The conversion formula is described in :cite:`Mietke2015`.

    Parameters
    ----------
    area_um: ndarray
        Convex cell area [µm²]
    deform: ndarray
        Deformation
    channel_width_in: float
        Original channel width [µm]
    channel_width_out: float
        Target channel width [µm]
    emodulus: ndarray
        Young's Modulus [kPa]
    flow_rate_in: float
        Original flow rate [µL/s]
    flow_rate_out: float
        Target flow rate [µL/s]
    viscosity_in: float
        Original viscosity [mPa*s]
    viscosity_out: float or ndarray
        Target viscosity [mPa*s]; This can be an array
    inplace: bool
        If True, override input arrays with corrected data

    Returns
    -------
    area_um_corr: ndarray
        Corrected cell area [µm²]
    deform_corr: ndarray
        Deformation (a copy if `inplace` is False)
    emodulus_corr: ndarray
        Corrected emodulus [kPa]; only returned if `emodulus` is given.

    Notes
    -----
    If only `area_um`, `deform`, `channel_width_in` and
    `channel_width_out` are given, then only the area is
    corrected and returned together with the original deform.
    If all other arguments are not set to None, the emodulus
    is corrected and returned as well.
    """
    warnings.warn("Usage of the `convert` method is deprecated! Please use "
                  + "the scale_feature method instead.",
                  DeprecationWarning)
    copy = not inplace

    deform_corr = np.array(deform, copy=copy)

    if channel_width_in != channel_width_out:
        area_um_corr = scale_area_um(area_um, channel_width_in,
                                     channel_width_out, inplace)
    else:
        area_um_corr = np.array(area_um, copy=copy)

    if (emodulus is not None
        and flow_rate_in is not None
        and flow_rate_out is not None
        and viscosity_in is not None
            and viscosity_out is not None):
        emodulus_corr = scale_emodulus(emodulus, channel_width_in,
                                       channel_width_out, flow_rate_in,
                                       flow_rate_out, viscosity_in,
                                       viscosity_out, inplace)

    if emodulus is None:
        return area_um_corr, deform_corr
    else:
        return area_um_corr, deform_corr, emodulus_corr


def scale_area_um(area_um, channel_width_in, channel_width_out, inplace=False,
                  **kwargs):
    """Perform scale conversion for area_um (linear elastic model)

    The area scales with the characteristic length
    "channel radius" L according to (L'/L)².

    The conversion formula is described in :cite:`Mietke2015`.

    .. versionadded: 0.25.0

    Parameters
    ----------
    area_um: ndarray
        Convex area [µm²]
    channel_width_in: float
        Original channel width [µm]
    channel_width_out: float
        Target channel width [µm]
    inplace: bool
        If True, override input arrays with corrected data
    kwargs:
        not used

    Returns
    -------
    area_um_corr: ndarray
        Scaled area [µm²]
    """
    copy = not inplace
    if issubclass(area_um.dtype.type, np.integer) and inplace:
        raise ValueError("Cannot correct integer `area_um` in-place!")
    area_um_corr = np.array(area_um, copy=copy)

    if channel_width_in != channel_width_out:
        area_um_corr *= (channel_width_out / channel_width_in)**2
    return area_um_corr


def scale_emodulus(emodulus, channel_width_in, channel_width_out,
                   flow_rate_in, flow_rate_out, viscosity_in,
                   viscosity_out, inplace=False):
    """Perform scale conversion for area_um (linear elastic model)

    The conversion formula is described in :cite:`Mietke2015`.

    .. versionadded: 0.25.0

    Parameters
    ----------
    emodulus: ndarray
        Young's Modulus [kPa]
    channel_width_in: float
        Original channel width [µm]
    channel_width_out: float
        Target channel width [µm]
    flow_rate_in: float
        Original flow rate [µL/s]
    flow_rate_out: float
        Target flow rate [µL/s]
    viscosity_in: float
        Original viscosity [mPa*s]
    viscosity_out: float or ndarray
        Target viscosity [mPa*s]; This can be an array
    inplace: bool
        If True, override input arrays with corrected data

    Returns
    -------
    emodulus_corr: ndarray
        Scaled emodulus [kPa]
    """
    copy = not inplace

    emodulus_corr = np.array(emodulus, copy=copy)

    if viscosity_in is not None:
        if isinstance(viscosity_in, np.ndarray):
            raise ValueError("`viscosity_in` must not be an array!")

    has_nones = (flow_rate_in is None
                 or flow_rate_out is None
                 or viscosity_in is None
                 or viscosity_out is None
                 or channel_width_in is None
                 or channel_width_out is None
                 )
    has_changes = (flow_rate_in != flow_rate_out
                   or channel_width_in != channel_width_out
                   or (isinstance(viscosity_out, np.ndarray)  # check before
                       or viscosity_in != viscosity_out)
                   )

    if not has_nones and has_changes:
        emodulus_corr *= (flow_rate_out / flow_rate_in) \
            * (viscosity_out / viscosity_in) \
            * (channel_width_in / channel_width_out)**3

    return emodulus_corr


def scale_feature(feat, data, inplace=False, **scale_kw):
    """Convenience function for scale conversions (linear elastic model)

    This method wraps around all the other scale_* methods and also
    supports deform/circ.

    Parameters
    ----------
    feat: str
        Valid scalar feature name
    data: float or ndarray
        Feature data
    inplace: bool
        If True, override input arrays with corrected data
    **scale_kw:
        Scale keyword arguments for the wrapped methods
    """
    if feat == "area_um":
        return scale_area_um(area_um=data, inplace=inplace, **scale_kw)
    elif feat in ["circ", "deform"]:
        # no units
        return np.array(data, copy=not inplace)
    elif feat == "emodulus":
        return scale_emodulus(emodulus=data, inplace=inplace, **scale_kw)
    elif feat == "volume":
        return scale_volume(volume=data, inplace=inplace, **scale_kw)
    else:
        raise KeyError("No recipe to scale feature '{}'!".format(feat))


def scale_volume(volume, channel_width_in, channel_width_out, inplace=False,
                 **kwargs):
    """Perform scale conversion for volume (linear elastic model)

    The volume scales with the characteristic length
    "channel radius" L according to (L'/L)³.

    .. versionadded: 0.25.0

    Parameters
    ----------
    volume: ndarray
        Volume [µm³]
    channel_width_in: float
        Original channel width [µm]
    channel_width_out: float
        Target channel width [µm]
    inplace: bool
        If True, override input arrays with corrected data
    kwargs:
        not used

    Returns
    -------
    volume_corr: ndarray
        Scaled volume [µm³]
    """
    copy = not inplace
    volume_corr = np.array(volume, copy=copy)

    if channel_width_in != channel_width_out:
        volume_corr *= (channel_width_out / channel_width_in)**3
    return volume_corr



"""Concatenate .rtdc files"""
import argparse
import time
import warnings

import numpy as np

from ..rtdc_dataset import new_dataset, RTDCWriter
from .. import definitions as dfn

from . import common


class FeatureSetNotIdenticalJoinWarning(UserWarning):
    pass


def join(path_out=None, paths_in=None, metadata=None):
    """Join multiple RT-DC measurements into a single .rtdc file"""
    if metadata is None:
        metadata = {"experiment": {"run index": 1}}
    if path_out is None or paths_in is None:
        parser = join_parser()
        args = parser.parse_args()
        paths_in = args.input
        path_out = args.output

    if len(paths_in) < 2:
        raise ValueError("At least two input files must be specified!")

    paths_in, path_out, path_temp = common.setup_task_paths(
        paths_in, path_out, allowed_input_suffixes=[".rtdc", ".tdms"])

    # Order input files by date
    key_paths = []
    for pp in paths_in:
        with new_dataset(pp) as ds:
            # sorting key
            key = "_".join([ds.config["experiment"]["date"],
                            ds.config["experiment"]["time"],
                            str(ds.config["experiment"]["run index"])
                            ])
            key_paths.append((key, pp))
    sorted_paths = [p[1] for p in sorted(key_paths, key=lambda x: x[0])]

    # Determine temporal offsets
    toffsets = np.zeros(len(sorted_paths), dtype=float)
    for ii, pp in enumerate(sorted_paths):
        with new_dataset(pp) as ds:
            etime = ds.config["experiment"]["time"]
            st = time.strptime(ds.config["experiment"]["date"]
                               + etime[:8],
                               "%Y-%m-%d%H:%M:%S")
            toffsets[ii] = time.mktime(st)
            if len(etime) > 8:
                # floating point time stored as well (HH:MM:SS.SS)
                toffsets[ii] += float(etime[8:])
    toffsets -= toffsets[0]

    logs = {}
    # Determine features to export (based on first file)
    with warnings.catch_warnings(record=True) as w:
        # Catch all FeatureSetNotIdenticalJoinWarnings
        warnings.simplefilter("ignore")
        warnings.simplefilter("always",
                              category=FeatureSetNotIdenticalJoinWarning)
        features = None
        for pp in sorted_paths:
            with new_dataset(pp) as ds:
                # features present
                if features is None:
                    # The initial features are the innate features of the
                    # first file (sorted by time). If we didn't use the innate
                    # features, then the resulting file might become large
                    # (e.g. if we included ancillary features).
                    features = sorted(ds.features_innate)
                else:
                    # Remove features from the feature list, if it is not in
                    # this dataset, or cannot be computed on-the-fly.
                    for feat in features:
                        if feat not in ds.features:
                            features.remove(feat)
                            warnings.warn(
                                f"Excluding feature '{feat}', because "
                                + f"it is not present in '{pp}'!",
                                FeatureSetNotIdenticalJoinWarning)
                    # Warn the user if this dataset has an innate feature that
                    # is being ignored, because it is not an innate feature of
                    # the first dataset.
                    for feat in ds.features_innate:
                        if feat not in features:
                            warnings.warn(
                                f"Ignoring feature '{feat}' in '{pp}', "
                                + "because it is not present in the "
                                + "other files being joined!",
                                FeatureSetNotIdenticalJoinWarning)
        if w:
            logs["dclab-join-feature-warnings"] = common.assemble_warnings(w)

    # Create initial output file
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        with new_dataset(sorted_paths[0]) as ds0:
            ds0.export.hdf5(path=path_temp,
                            features=features,
                            filtered=False,
                            override=True,
                            compression="gzip")
        if w:
            logs["dclab-join-warnings-#1"] = common.assemble_warnings(w)

    with RTDCWriter(path_temp) as hw:
        ii = 1
        # Append data from other files
        for pi, ti in zip(sorted_paths[1:], toffsets[1:]):
            ii += 1  # we start with the second dataset
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                with new_dataset(pi) as dsi:
                    for feat in features:
                        if feat == "time":
                            # handle time offset
                            fdata = dsi["time"] + ti
                        elif feat == "frame":
                            # handle frame offset
                            fr = dsi.config["imaging"]["frame rate"]
                            frame_offset = ti * fr
                            fdata = dsi["frame"] + frame_offset
                        elif feat == "index_online":
                            if "events/index_online" in hw.h5file:
                                # index_online is usually larger than index
                                ido0 = hw.h5file["events/index_online"][-1] + 1
                            else:
                                ido0 = 0
                            fdata = dsi["index_online"] + ido0
                        else:
                            fdata = dsi[feat]
                        hw.store_feature(feat=feat, data=fdata)
                if w:
                    lkey = f"dclab-join-warnings-#{ii}"
                    logs[lkey] = common.assemble_warnings(w)

        # Logs and configs from source files
        logs["dclab-join"] = common.get_command_log(paths=sorted_paths)
        for ii, pp in enumerate(sorted_paths):
            with new_dataset(pp) as ds:
                # data file logs
                for ll in ds.logs:
                    logs[f"src-#{ii+1}_{ll}"] = ds.logs[ll]
                # configuration
                cfg = ds.config.tostring(sections=dfn.CFG_METADATA).split("\n")
                logs[f"cfg-#{ii+1}"] = cfg

        # Write logs and missing meta data
        for name in logs:
            hw.store_log(name, logs[name])
        hw.store_metadata(metadata)

    # Finally, rename temp to out
    path_temp.rename(path_out)


def join_parser():
    descr = "Join two or more RT-DC measurements. This will produce " \
            + "one larger .rtdc file. The meta data of the dataset " \
            + "that was recorded earliest will be used in the output " \
            + "file. Please only join datasets that were recorded " \
            + "in the same measurement run."
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('input', metavar="INPUT", nargs="*", type=str,
                        help='Input paths (.tdms or .rtdc files)')
    required_named = parser.add_argument_group('required named arguments')
    required_named.add_argument('-o', '--output', metavar="OUTPUT", type=str,
                                help='Output path (.rtdc file)', required=True)
    return parser



import logging
import os
import subprocess as sp

from .rlibs import (
    RUnavailableError, rpy2, rpy2_is_version_3, import_r_submodules)

# Disable rpy2 logger because of unnecessary prints to stdout
logging.getLogger("rpy2.rinterface_lib.callbacks").disabled = True


class RNotFoundError(BaseException):
    pass


class AutoRConsole(object):
    """Helper class for catching R console output"""
    lock = False
    perform_lock = rpy2_is_version_3

    def __init__(self):
        """
        By default, this console always returns "yes" when asked a
        question. If you need something different, you can subclass
        and override `consoleread` fucntion. The console stream is
        recorded in `self.stream`.
        """
        self.stream = [["init", "Starting RConsole class\n"]]
        if AutoRConsole.perform_lock:
            if AutoRConsole.lock:
                raise ValueError("Only one RConsole instance allowed!")
            AutoRConsole.lock = True
            self.original_funcs = {
                "consoleread": rpy2.rinterface_lib.callbacks.consoleread,
                "consolewrite_print":
                    rpy2.rinterface_lib.callbacks.consolewrite_print,
                "consolewrite_warnerror":
                    rpy2.rinterface_lib.callbacks.consolewrite_warnerror,
            }
            rpy2.rinterface_lib.callbacks.consoleread = self.consoleread
            rpy2.rinterface_lib.callbacks.consolewrite_print = \
                self.consolewrite_print
            rpy2.rinterface_lib.callbacks.showmessage = \
                self.consolewrite_print

            rpy2.rinterface_lib.callbacks.consolewrite_warnerror = \
                self.consolewrite_warnerror
        # Set locale (to get always English messages)
        rpy2.robjects.r('Sys.setlocale("LC_MESSAGES", "C")')
        rpy2.robjects.r('Sys.setlocale("LC_CTYPE", "C")')

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if AutoRConsole.perform_lock:
            AutoRConsole.lock = False
            rpy2.rinterface_lib.callbacks.consoleread = \
                self.original_funcs["consoleread"]
            rpy2.rinterface_lib.callbacks.consolewrite_print = \
                self.original_funcs["consolewrite_print"]
            rpy2.rinterface_lib.callbacks.consolewrite_warnerror = \
                self.original_funcs["consolewrite_warnerror"]

    def close(self):
        """Remove the rpy2 monkeypatches"""
        self.__exit__()

    def consoleread(self, prompt):
        """Read user input, returns "yes" by default"""
        self.write_to_stream("consoleread", prompt + "YES")
        return "yes"

    def consolewrite_print(self, s):
        self.write_to_stream("consolewrite_print", s)

    def consolewrite_warnerror(self, s):
        self.write_to_stream("consolewrite_warnerror", s)

    def write_to_stream(self, topic, s):
        prev_topic = self.stream[-1][0]
        same_topic = prev_topic == topic
        unfinished_line = self.stream[-1][1][-1] not in ["\n", "\r"]
        if same_topic and unfinished_line:
            # append to previous line
            self.stream[-1][1] += s
        else:
            self.stream.append([topic, s])

    def get_prints(self):
        prints = []
        for line in self.stream:
            if line[0] == "consolewrite_print":
                prints.append(line[1].strip())
        return prints

    def get_warnerrors(self):
        warnerrors = []
        for line in self.stream:
            if line[0] == "consolewrite_warnerror":
                warnerrors.append(line[1].strip())
        return warnerrors


def check_r():
    """Make sure R is installed an R HOME is set"""
    if not has_r():
        raise RNotFoundError("Cannot find R, please set its path with the "
                             + "`set_r_path` function.")


def get_r_path():
    """Get the path of the R executable/binary from rpy2"""
    r_home = rpy2.situation.get_r_home()
    return rpy2.situation.get_r_exec(r_home)


def get_r_version():
    check_r()
    ver_string = rpy2.situation.r_version_from_subprocess().strip()
    if ver_string:
        # get the actual version string
        if ver_string.startswith("R version "):
            ver_string = ver_string.split(" ")[2]
    return ver_string


def has_lme4():
    """Return True if the lme4 package is installed"""
    check_r()
    lme4_there = rpy2.robjects.packages.isinstalled("lme4")
    statmod_there = rpy2.robjects.packages.isinstalled("statmod")
    return lme4_there and statmod_there


def has_r():
    """Return True if R is available"""
    try:
        hasr = rpy2.situation.get_r_home() is not None
    except RUnavailableError:
        hasr = False
    return hasr


def import_lme4():
    check_r()
    if has_lme4():
        lme4pkg = rpy2.robjects.packages.importr("lme4")
    else:
        raise ValueError(
            "The R package 'lme4' is not installed, please install it via "
            + "`dclab.lme4.rsetup.install_lme4()` or by executing "
            + "in a shell: R -e " + '"install.packages(' + "'lme4', "
            + "repos='http://cran.r-project.org')" + '"')
    return lme4pkg


def install_lme4():
    """Install the lme4 package (if not already installed)

    The packages are installed to the user data directory
    given in :const:`lib_path`.
    """
    check_r()
    if not has_lme4():
        # import R's utility package
        utils = rpy2.robjects.packages.importr('utils')
        # select the first mirror in the list
        utils.chooseCRANmirror(ind=1)
        # install lme4 to user data directory (say yes to user dir install)
        with AutoRConsole() as rc:
            # install statmod first (Doesn't R have package dependencies?!)
            utils.install_packages(
                rpy2.robjects.vectors.StrVector(["statmod", "lme4"]))
        return rc


def set_r_path(r_path):
    """Set the path of the R executable/binary for rpy2"""
    tmp = sp.check_output((r_path, 'RHOME'), universal_newlines=True)
    r_home = tmp.split(os.linesep)
    if r_home[0].startswith('WARNING'):
        res = r_home[1]
    else:
        res = r_home[0].strip()
    os.environ["R_HOME"] = res
    import_r_submodules()



from ..rtdc_dataset.feat_anc_core.ancillary_feature import AncillaryFeature

from . import feat_const


def check_feature_shape(name, data):
    """Check if (non)-scalar feature matches with its data's dimensionality

    Parameters
    ----------
    name: str
        name of the feature
    data: array-like
        data whose dimensionality will be checked

    Raises
    ------
    ValueError
        If the data's shape does not match its scalar description
    """
    if len(data.shape) == 1 and not scalar_feature_exists(name):
        raise ValueError(f"Feature '{name}' is not a scalar feature, but "
                         "a 1D array was given for `data`!")
    elif len(data.shape) != 1 and scalar_feature_exists(name):
        raise ValueError(f"Feature '{name}' is a scalar feature, but the "
                         "`data` array is not 1D!")


def feature_exists(name, scalar_only=False):
    """Return True if `name` is a valid feature name

    This function not only checks whether `name` is in
    :const:`feature_names`, but also validates against
    the machine learning scores `ml_score_???` (where
    `?` can be a digit or a lower-case letter in the
    English alphabet).

    Parameters
    ----------
    name: str
        name of a feature
    scalar_only : bool
        Specify whether the check should only search in scalar features

    Returns
    -------
    valid: bool
        True if name is a valid feature, False otherwise.

    See Also
    --------
    scalar_feature_exists: Wraps `feature_exists` with `scalar_only=True`
    """
    valid = False
    if name in feat_const.scalar_feature_names:
        # scalar feature
        valid = True
    elif not scalar_only and name in feat_const.feature_names:
        # non-scalar feature
        valid = True
    else:
        # check whether we have an `ml_score_???` feature
        valid_chars = "0123456789abcdefghijklmnopqrstuvwxyz"
        if (name.startswith("ml_score_")
            and len(name) == len("ml_score_???")
            and name[-3] in valid_chars
            and name[-2] in valid_chars
                and name[-1] in valid_chars):
            valid = True
    return valid


def feature_register(name, label=None, is_scalar=True):
    """Register a new feature for usage in dclab

    Used by temporary features and plugin features to add new feature
    names and labels to `dclab.definitions`.

    Parameters
    ----------
    name: str
        name of a feature
    label: str, optional
        feature label corresponding to the feature name. If set to None, then
        a label is constructed for the feature name.
    is_scalar: bool
        Specify whether the feature of an event is a scalar (True)
        or not (False)

    Raises
    ------
    ValueError
        If the feature already exists.
    """
    allowed_chars = "abcdefghijklmnopqrstuvwxyz_1234567890"
    feat = "".join([f for f in name if f in allowed_chars])
    if feat != name:
        raise ValueError("`feature` must only contain lower-case characters, "
                         f"digits, and underscores; got '{name}'!")
    if label is None:
        label = f"User-defined feature {name}"
    if feature_exists(name):
        raise ValueError(f"Feature '{name}' already exists!")

    # Populate the new feature in all dictionaries and lists
    # (we don't need global here)
    feat_const.feature_names.append(name)
    feat_const.feature_labels.append(label)
    feat_const.feature_name2label[name] = label
    if is_scalar:
        feat_const.scalar_feature_names.append(name)


def feature_deregister(name):
    """Unregister a feature from dclab

    Used by temporary features and plugin features to
    remove the feature names and labels from `dclab.definitions`.

    Parameters
    ----------
    name: str
        name of a feature

    Warnings
    --------
    This function should only be used internally, i.e., You should not use
    this function. This function can break things.
    """
    label = get_feature_label(name)
    feat_const.feature_names.remove(name)
    feat_const.feature_labels.remove(label)
    feat_const.feature_name2label.pop(name)
    if name in feat_const.scalar_feature_names:
        feat_const.scalar_feature_names.remove(name)


def get_feature_label(name, rtdc_ds=None, with_unit=True):
    """Return the label corresponding to a feature name

    This function not only checks :const:`feature_name2label`,
    but also supports registered `ml_score_???` features.

    Parameters
    ----------
    name: str
        name of a feature
    with_unit: bool
        set to False to remove units in square brackets

    Returns
    -------
    label: str
        feature label corresponding to the feature name

    Notes
    -----
    TODO: extract feature label from ancillary information when an rtdc_ds is
    given.
    """
    assert feature_exists(name)
    if name in feat_const.feature_name2label:
        label = feat_const.feature_name2label[name]
    else:
        # First check whether an ancillary feature with that
        # name exists.
        for af in AncillaryFeature.features:
            if af.feature_name == name:
                labelid = af.data.outputs.index(name)
                label = af.data.output_labels[labelid]
                break
        else:
            # If that did not work, use a generic name.
            label = "ML score {}".format(name[-3:].upper())
    if not with_unit:
        if label.endswith("]") and label.count("["):
            label = label.rsplit("[", 1)[0].strip()
    return label


def scalar_feature_exists(name):
    """Convenience method wrapping `feature_exists(..., scalar_only=True)`"""
    return feature_exists(name, scalar_only=True)




import numpy as np
from .ancillary_feature import AncillaryFeature


def get_ml_score_names(mm):
    """Return a list of all ml_score_??? features"""
    feats = []
    # We cannot loop over mm.features because of infinite recursions
    for ft in mm._feature_candidates:
        if ft.startswith("ml_score_") and ft in mm:
            feats.append(ft)
    return sorted(feats)


def compute_ml_class(mm, sanity_checks=True):
    """Compute the most-probable class membership for all events

    Parameters
    ----------
    mm: .RTDCBase
        instance with the `ml_score_???` features
    sanity_checks: bool
        set to `False` to not perform sanity checks (checks whether
        the scores are between 0 and 1)

    Returns
    -------
    ml_class: 1D ndarray
        The most-probable class for each event in `mm`. If no class
        can be attributed to an event (because the scores are all
        `np.nan` or `0` for that event), the class `-1` is used.

    Notes
    -----
    I initially thought about also checking whether each feature
    sums to one, but discarded the idea. Let's assume that a classifier
    does an awful classification and classifies all events in
    the same way. If the dataset is cropped at some point (e.g.
    debris or other events), then this bad classifier has an
    increased probability compared to another classifier which is
    perfect at picking out one population. The ml_score values
    should be just in the range of [0, 1]. This also simplifies
    export to hdf5 and the work with hierarchy children.
    """
    feats = get_ml_score_names(mm)

    # the score matrix
    score_matrix = np.zeros((len(mm), len(feats)), dtype=float)

    for ii, ft in enumerate(feats):
        if sanity_checks:
            if np.nanmax(mm[ft]) > 1:
                raise ValueError("Feature '{}' has values > 1!".format(ft))
            elif np.nanmin(mm[ft]) < 0:
                raise ValueError("Feature '{}' has values < 0!".format(ft))
        score_matrix[:, ii] = mm[ft]

    # Now compute the maximum for each event. The initial idea was to just
    # use `ml_class = np.nanargmax(score_matrix, axis=1)`. However, here we
    # run into these problems:
    # 1. This does not handle All-NaN slices, e.g. all features are `np.nan`
    #    for an event.
    # 2. This does not properly handle manually-rated, zero-valued features,
    #    e.g. in a situation where we have two features, one with `np.nan`
    #    and one with `0`, we cannot assign the event to either of the two
    #    classes.
    # 3. There is no "unclassified" class (this also becomes apparent in
    #    point 2). We will set all events that cannot be attributed to a
    #    class to `-1` in `ml_class`.

    # Define unusable entries:
    unusable = np.logical_or(np.isnan(score_matrix), (score_matrix == 0))
    where_idx_nan = np.sum(~unusable, axis=1) == 0
    score_matrix[where_idx_nan, :] = -1
    ml_class = np.nanargmax(score_matrix, axis=1)
    ml_class[where_idx_nan] = -1
    return ml_class


def has_ml_scores(mm):
    """Check whether the dataset has ml_scores defined"""
    # Return the sorted score names plus Ancillary feature hashes.
    # This will be used to determine the hash of the ml_class feature,
    # which is important in case the user replaces an ML feature
    # with a new one.
    features = get_ml_score_names(mm)
    idlist = []
    for feat in features:
        # We also hash any other AncillaryFeature that might implement
        # this ML score. But this use case is basically non-existent and
        # the performance impact is probably negligible.
        candidates = AncillaryFeature.get_instances(feat)
        idlist.append((feat, [c.hash(mm) for c in candidates]))
    return idlist


def register():
    AncillaryFeature(feature_name="ml_class",
                     method=compute_ml_class,
                     req_func=has_ml_scores,
                     )



def compute_my_feature(rtdc_ds):
    """Compute circularity times area"""
    circ_times_area = rtdc_ds["circ"] * rtdc_ds["area_um"]
    return {"circ_times_area": circ_times_area}


info = {
    "method": compute_my_feature,
    "description": "Compute area times circularity",
    "feature names": ["circ_times_area"],
    "features required": ["circ", "area_um"],
    "version": "0.1.0",
}
